from __future__ import annotations
from abc import abstractmethod
from typing import Any
from equinox import filter
from jax import Array, numpy as np
import jax.tree as jtu
import dLux.utils as dlu

from ..wavefronts import Wavefront
from ..transformations import CoordTransform
from .optical_layers import OpticalLayer, BasisLayer
from .aberrations import ZernikeBasis


__all__ = [
    "CircularAperture",
    "RectangularAperture",
    "RegPolyAperture",
    "Spider",
    "SquareAperture",
    "AberratedAperture",
    "CompoundAperture",
    "MultiAperture",
]


class ApertureLayer(OpticalLayer):
    """
    Base ApertureLayer class, instantiates the normalise attribute.

    Attributes
    ----------
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    normalise: bool

    def __init__(self: OpticalLayer, normalise: bool = False, **kwargs):
        """
        Parameters
        ----------
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        self.normalise = bool(normalise)
        super().__init__(**kwargs)

    @abstractmethod
    def transmission(self, coords, pixel_scale):  # pragma: no cover
        """
        Evaluates the aperture transmission on the given coords, applying the aperture
        transformations to the coords.
        """


class BaseDynamicAperture(ApertureLayer):
    """
    An abstract base class that implements the coordinate transformation attribute.

    Attributes
    ----------
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    transformation: CoordTransform

    def __init__(
        self: ApertureLayer,
        transformation: CoordTransform = None,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        normalise : bool = False
            Should the aperture normalise the wavefront after being applied.
        """
        super().__init__(normalise=normalise)
        if transformation is not None:
            if not isinstance(transformation, CoordTransform):
                raise TypeError(
                    "transformation must be a CoordTransform object. "
                    "Use the CoordTransform class to create one."
                )
        self.transformation = transformation

    def __getattr__(self: ApertureLayer, key: str) -> Any:
        """Raises transformation attributes to the ApertureLayer level."""
        if hasattr(self.transformation, key):
            return getattr(self.transformation, key)
        raise AttributeError(f"{key} not in {self.__class__.__name__}")

    def apply(self: ApertureLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        # Apply aperture
        wavefront *= self.transmission(
            wavefront.coordinates, wavefront.pixel_scale
        )
        if self.normalise:
            return wavefront.normalise()
        return wavefront


class DynamicAperture(BaseDynamicAperture):
    """
    An abstract base class that implements the methods required to provide dynamic soft
    edges to the apertures, and the ability to occult the aperture.

    Attributes
    ----------
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
        transmissive aperture, and True results in an occulting aperture.
    softness: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    occulting: bool
    softness: float

    def __init__(
        self: ApertureLayer,
        transformation: CoordTransform = None,
        occulting: bool = False,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        occulting: bool = False
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: float, pixels = 1.0
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        super().__init__(
            transformation=transformation,
            normalise=normalise,
        )
        self.occulting = bool(occulting)
        self.softness = float(softening)
        if self.softness <= 0:
            raise ValueError("softening must be greater than 0.")

    @abstractmethod  # pragma: no cover
    def extent(self):
        pass


#############################
# Concrete Aperture Classes #
#############################
class CircularAperture(DynamicAperture):
    """
    A dynamically generated circular aperture parameterised by its radius. Both jit and
    grad compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/CircularAperture.png)

    Attributes
    ----------
    radius: float, meters
        The radius of the aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
    softening: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    radius: float

    def __init__(
        self: ApertureLayer,
        radius: float,
        transformation: CoordTransform = None,
        occulting: bool = False,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        radius: Array, meters
            The radius of the aperture.

        occulting: bool = False
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: Array, pixels = np.array(1.)
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        super().__init__(
            transformation=transformation,
            occulting=occulting,
            softening=softening,
            normalise=normalise,
        )
        self.radius = float(radius)

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        clip_val = pixel_scale * self.softness / 2
        return dlu.soft_circle(coords, self.radius, clip_val, self.occulting)

    @property
    def extent(self: ApertureLayer) -> float:
        """
        Returns the maximum extent of the aperture.

        Returns
        -------
        extent : float
            The maximum extent of the aperture.
        """
        return self.radius

    @property
    def nsides(self: ApertureLayer) -> int:
        """
        Returns the number of sides of the aperture.

        Returns
        -------
        nsides : int
            The number of sides of the aperture.
        """
        return 0


class SquareAperture(DynamicAperture):
    """
    A dynamically generated square aperture parameterised by it side length.
    Both jit and grad compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/SquareAperture.png)

    Attributes
    ----------
    width: float, meters
        The side length of the aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
        transmissive aperture, and True results in an occulting aperture.
    softening: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    width: float

    def __init__(
        self: ApertureLayer,
        width: float,
        transformation: CoordTransform = None,
        occulting: bool = False,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        width: Array, meters
            The side length of the aperture.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        occulting: bool = False
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: float, pixels = 1.0
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        super().__init__(
            transformation=transformation,
            occulting=occulting,
            softening=softening,
            normalise=normalise,
        )

        self.width = float(width)

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        clip_val = pixel_scale * self.softness / 2
        return dlu.soft_square(coords, self.width, clip_val, self.occulting)

    @property
    def extent(self: ApertureLayer) -> float:
        """
        Returns the maximum extent of the aperture.

        Returns
        -------
        extent : float
            The maximum extent of the aperture.
        """
        return np.sqrt(2) * self.width

    @property
    def nsides(self: ApertureLayer) -> int:
        """
        Returns the number of sides of the aperture.

        Returns
        -------
        nsides : int
            The number of sides of the aperture.
        """
        return 4


class RectangularAperture(DynamicAperture):
    """
    A dynamically generated rectangular aperture parameterised by it width and height.
    Both jit and grad compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/RectangularAperture.png)

    Attributes
    ----------
    height: float, meters
        The length of the aperture in the y-direction.
    width: float, meters
        The length of the aperture in the x-direction.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
        transmissive aperture, and True results in an occulting aperture.
    softening: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    height: float
    width: float

    def __init__(
        self: ApertureLayer,
        height: float,
        width: float,
        transformation: CoordTransform = None,
        occulting: bool = False,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        height: Array, meters
            The length of the aperture in the y-direction.
        width: Array, meters
            The length of the aperture in the x-direction.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        occulting: bool = False
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: float, pixels = 1.0
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        self.height = float(height)
        self.width = float(width)

        super().__init__(
            transformation=transformation,
            occulting=occulting,
            softening=softening,
            normalise=normalise,
        )

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        clip_val = pixel_scale * self.softness / 2
        return dlu.soft_rectangle(
            coords, self.width, self.height, clip_val, self.occulting
        )

    @property
    def extent(self: ApertureLayer) -> float:
        """
        Returns the maximum extent of the aperture.

        Returns
        -------
        extent : float
            The maximum extent of the aperture.
        """
        return np.hypot(self.height / 2.0, self.width / 2.0)

    @property
    def nsides(self: ApertureLayer) -> int:
        """
        Returns the number of sides of the aperture.

        Returns
        -------
        nsides : int
            The number of sides of the aperture.
        """
        return 4


class RegPolyAperture(DynamicAperture):
    """
    Creates a dynamically generated regular polygon aperture parameterised by its
    number of sides and maximum radius. Both jit and grad compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/RegPolyAperture.png)

    Attributes
    ----------
    nsides: int
        The number of sides of the aperture.
    rmax: float, meters
        The maximum radius to the vertices from its center.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
        transmissive aperture, and True results in an occulting aperture.
    softening: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    nsides: int
    rmax: float

    def __init__(
        self: ApertureLayer,
        nsides: int,
        rmax: float,
        transformation: CoordTransform = None,
        occulting: bool = False,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        nsides: int
            The number of sides of the aperture.
        rmax: float, meters
            The maximum radius to the vertices from its center.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        occulting: bool = False
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: float, pixels = 1.0
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        self.nsides = int(nsides)
        self.rmax = float(rmax)

        super().__init__(
            transformation=transformation,
            occulting=occulting,
            softening=softening,
            normalise=normalise,
        )

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        clip_val = pixel_scale * self.softness / 2
        return dlu.soft_reg_polygon(
            coords, self.rmax, self.nsides, clip_val, self.occulting
        )

    @property
    def extent(self: ApertureLayer) -> float:
        """
        Returns the maximum extent of the aperture.

        Returns
        -------
        extent : float
            The maximum extent of the aperture.
        """
        return self.rmax


class Spider(DynamicAperture):
    """
    Creates a dynamically generated spider aperture parameterised by its arm width and
    number of arms. Both jit and grad compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/Spider.png)

    Attributes
    ----------
    width: float, meters
        The width of the spider.
    angles: Array, degrees
        The angle of each arm of the spider.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    occulting: bool
        Is the aperture occulting or transmissive. False results in a
        transmissive aperture, and True results in an occulting aperture.
    softening: float, pixels
        The approximate pixel width of the soft boundary applied to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    width: float
    angles: Array

    def __init__(
        self: ApertureLayer,
        width: float,
        angles: Array,
        transformation: CoordTransform = None,
        occulting: bool = True,
        softening: float = 1.0,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        width: float, meters
            The width of the spider.
        angles: Array, degrees
            The angle of each arm of the spider.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        occulting: bool = True
            Is the aperture occulting or transmissive. False results in a
            transmissive aperture, and True results in an occulting aperture.
        softening: float, pixels = 1.0
            The approximate pixel width of the soft boundary applied to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        super().__init__(
            transformation=transformation,
            occulting=occulting,
            softening=softening,
            normalise=normalise,
        )

        self.width = float(width)
        self.angles = np.asarray(angles, dtype=float)

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        clip_val = pixel_scale * self.softness / 2
        return dlu.soft_spider(
            coords, self.width, self.angles, clip_val, self.occulting
        )

    @property
    def extent(self: ApertureLayer) -> float:
        """
        Returns the maximum extent of the aperture.

        Returns
        -------
        extent : float
            The maximum extent of the aperture.
        """
        raise TypeError("Spiders do not have an extent.")

    @property
    def nsides(self: ApertureLayer) -> int:
        """
        Returns the number of sides of the aperture.

        Returns
        -------
        nsides : int
            The number of sides of the aperture.
        """
        raise TypeError("Spiders do not have a number of sides.")


# TODO: Should this not have transmission as it is held by its sub-aperture?
class AberratedAperture(BasisLayer, ApertureLayer):
    """
    Creates a dynamically generated Aperture with aberrations. Both jit and grad
    compatible.

    ??? abstract "UML"
        ![UML](../../assets/uml/AberratedAperture.png)

    Attributes
    ----------
    aperture: ApertureLayer
        The aperture on which the aberration basis is defined.
    basis: list[Zernike]
        A list of basis functions that generate the basis vectors.
    coefficients: Array
        The amplitude of each basis vector of the aberrations.
    as_phase : bool
        Whether to apply the basis as a phase phase or OPD. If True the basis is
        applied as a phase, else it is applied as an OPD.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    aperture: ApertureLayer

    def __init__(
        self: ApertureLayer,
        aperture: ApertureLayer,
        noll_inds: Array,
        coefficients: Array = None,
        as_phase: bool = False,
    ):
        """
        Parameters
        ----------
        aperture: ApertureLayer
            The aperture on which the aberration basis is defined.
        noll_inds: Array
            The noll indices of the basis functions to use.
        coefficients: Array = None
            The amplitude of each basis vector of the aberrations.
        as_phase : bool = False
            Whether to apply the basis as a phase phase or OPD. If True the basis is
            applied as a phase, else it is applied as an OPD.
        """
        # Ensure aperture is dynamic
        if not isinstance(aperture, DynamicAperture):
            raise TypeError(
                "AberratedApertures can not contain Static, "
                "Compound or Multi Apertures. AberratedApertures can be "
                "placed in Compound or Multi Apertures, which can then be "
                "promoted to Static."
            )

        super().__init__(normalise=aperture.normalise, as_phase=as_phase)

        # Ensure transmissive
        if aperture.occulting:
            raise TypeError("AberratedApertures can not be occulting.")

        if isinstance(aperture, Spider):
            raise TypeError("AberratedApertures can not be spiders.")

        # Set Aperture
        self.aperture = aperture
        self.basis = ZernikeBasis(noll_inds)

        if coefficients is None:
            coefficients = np.zeros(len(noll_inds))
        self.coefficients = np.asarray(coefficients, dtype=float)

    def calculate(self: ApertureLayer):
        """
        Required abstract method. Raises NotImplementedError as it is invalid here.
        """
        raise NotImplementedError(
            "Aberrated Apertures can not use the .calculate() method because "
            "they need coords to be generated on. please use "
            ".eval_basis(coords) method instead."
        )

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        return self.aperture.transmission(coords, pixel_scale)

    def calc_basis(self: ApertureLayer, coords: Array) -> Array:
        """
        Calculates the basis vectors at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the basis vectors on.

        Returns
        -------
        basis : Array
            The basis vectors at the given coordinates.
        """
        if self.aperture.transformation is not None:
            coords = self.aperture.transformation.apply(coords)
        coords /= self.aperture.extent
        return self.basis.calculate_basis(coords, self.aperture.nsides)

    def eval_basis(self: ApertureLayer, coords: Array) -> Array:
        """
        Evaluates the basis vectors at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to evaluate the basis vectors on.

        Returns
        -------
        aberrations : Array
            The aberrations at the given coordinates.
        """
        basis = self.calc_basis(coords)
        return dlu.eval_basis(basis, self.coefficients)

    def apply(self: ApertureLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        # Transmission
        wavefront *= self.transmission(
            wavefront.coordinates, wavefront.pixel_scale
        )
        if self.normalise:
            wavefront = wavefront.normalise()

        # Transform coordinate
        if self.aperture.transformation is not None:
            coords = self.aperture.transformation.apply(wavefront.coordinates)
        else:
            coords = wavefront.coordinates

        # Aberrations
        aberrations = self.eval_basis(coords)
        if self.as_phase:
            wavefront = wavefront.add_phase(aberrations)
        else:
            wavefront += aberrations
        return wavefront


class CompositeAperture(BaseDynamicAperture):
    """
    Base class for dynamically generated Apertures that are a combination of various
    different aperture shapes. Both jit and grad compatible.

    Attributes
    ----------
    apertures: dict
       The sub-apertures that make up the full aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    apertures: dict

    def __init__(
        self: ApertureLayer,
        apertures: list,
        transformation: CoordTransform = None,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        apertures: list[ApertureLayer]
            The sub-apertures that make up the full aperture.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        self.apertures = dlu.list2dictionary(apertures, False, ApertureLayer)
        super().__init__(
            transformation=transformation,
            normalise=normalise,
        )

    def __getattr__(self: ApertureLayer, key: str) -> Any:
        """
        Raises the contained apertures via their dictionary keys.

        Parameters
        ----------
        key: str
            The attribute to get.

        Returns
        -------
        attribute: Any
            The aperture found at the given key.
        """
        if key in list(self.apertures.keys()):
            return self.apertures[key]
        else:
            raise AttributeError(f"{key} not in {self.apertures.keys()}")

    def _aberrated_apertures(self: ApertureLayer) -> list[ApertureLayer]:
        """
        Returns the list of aberrated apertures, from the full set of apertures.

        Returns
        -------
        apertures : list
            The list of aberrated apertures.
        """

        # Define leaf fn
        def is_aberrated(leaf):
            if isinstance(leaf, AberratedAperture):
                return True
            elif isinstance(leaf, CompoundAperture):
                if len(leaf._aberrated_apertures()) > 0:
                    return True
            return False

        # Get aberrated apertures
        # TODO: use partition
        filter_map = jtu.map(
            is_aberrated, self.apertures, is_leaf=is_aberrated
        )
        aberrated = filter(self.apertures, filter_map)
        return jtu.flatten(aberrated, is_leaf=is_aberrated)[0]

    @property
    def coefficients(self: ApertureLayer) -> list[Array]:
        """
        Returns the coefficients of the aberrated apertures.

        Returns
        -------
        coefficients : list[Array]
            The coefficients of the aberrated apertures.
        """
        return [ap.coefficients for ap in self._aberrated_apertures()]

    def calc_basis(self: ApertureLayer, coords: Array) -> Array:
        """
        Calculates the basis vectors at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the basis vectors on.

        Returns
        -------
        basis : Array
            The basis vectors at the given coordinates.
        """
        aberrated_apertures = self._aberrated_apertures()
        basis_fn = lambda ap: ap.calc_basis(coords)
        leaf_fn = lambda ap: isinstance(ap, AberratedAperture)
        basii = jtu.map(basis_fn, aberrated_apertures, is_leaf=leaf_fn)
        return jtu.flatten(basii)[0]

    def eval_basis(self: ApertureLayer, coords: Array) -> Array:
        """
        Evaluates the basis vectors at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to evaluate the basis vectors on.

        Returns
        -------
        aberrations : Array
            The aberrations at the given coordinates.
        """
        basii = self.calc_basis(coords)
        coeffs = self.coefficients
        eval_fn = lambda basis, coeff: dlu.eval_basis(basis, coeff)
        return np.array(jtu.map(eval_fn, basii, coeffs))

    def transmissions(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        eval_fn = lambda ap: ap.transmission(coords, pixel_scale)
        leaf_fn = lambda ap: isinstance(ap, ApertureLayer)
        transmissions = jtu.map(eval_fn, self.apertures, is_leaf=leaf_fn)
        return np.squeeze(np.array(jtu.flatten(transmissions)[0]))

    def apply(self: ApertureLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        # Transmission
        wavefront *= self.transmission(
            wavefront.coordinates, wavefront.pixel_scale
        )
        if self.normalise:
            return wavefront.normalise()

        # Transform coordinate
        if self.transformation is not None:
            coords = self.transformation.apply(wavefront.coordinates)
        else:
            coords = wavefront.coordinates

        # Aberrations
        aberrations = self.eval_basis(coords)
        # TODO: Add add_phase as an option
        # if self.as_phase:
        #     wavefront = wavefront.add_phase(aberrations)
        # else:
        wavefront += aberrations
        return wavefront


class CompoundAperture(CompositeAperture):
    """
    Dynamically generates an Apertures from a series of overlapping sub-apertures. Both
    jit and grad compatible.

    This class combines the aperture via a _multiplication_ of the sub-apertures. An
    example would be a HST-like aperture with an obscuring secondary mirror and spiders.

    This class can only contain a _single_ AberratedAperture.

    If you want to combine apertures via an _addition_ of the sub-apertures such as an
    aperture mask, use the MultiAperture class.

    Note that this class can not contain a MultiAperture, but MultiApertures can
    contain CompoundApertures.

    ??? abstract "UML"
        ![UML](../../assets/uml/CompoundAperture.png)

    Attributes
    ----------
    apertures: dict
        The sub-apertures that make up the full aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    def __init__(
        self: ApertureLayer,
        apertures: list,
        transformation: CoordTransform = None,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        apertures: list[ApertureLayer]
            The sub-apertures that make up the full aperture.
        transformation: CoordTransform
            The object that applies the coordinate transformations to the aperture.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the aperture.
        """
        super().__init__(
            apertures=apertures,
            transformation=transformation,
            normalise=normalise,
        )

        # Check for more than one aberrated aperture
        naberrated = 0
        for aperture in self.apertures.values():
            if isinstance(aperture, AberratedAperture):
                naberrated += 1
            if naberrated > 1:
                raise TypeError(
                    "CompoundAperture can only have a single "
                    "AberratedAperture."
                )

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        return self.transmissions(coords, pixel_scale).prod(0)


class MultiAperture(CompositeAperture):
    """
    Dynamically generates an Apertures from a series of separated sub-apertures. Both
    jit and grad compatible.

    This class combines the aperture via an _addition_ of the sub-apertures. An example
    would be a aperture mask with multiple holes.

    This class can only contain a _multiple_ AberratedAperture, or CompoundApertures.

    If you want to combine apertures via a _multiplication_ of the sub-apertures such
    as HST-like aperture, use the CompoundAperture class.

    ??? abstract "UML"
        ![UML](../../assets/uml/MultiAperture.png)

    Attributes
    ----------
    apertures: dict
        The sub-apertures that make up the full aperture.
    transformation: CoordTransform
        The object that applies the coordinate transformations to the aperture.
    normalise : bool
        Whether to normalise the wavefront after passing through the aperture.
    """

    def eval_basis(self: ApertureLayer, coords: Array) -> Array:
        """
        Evaluates the basis vectors at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to evaluate the basis vectors on.

        Returns
        -------
        aberrations : Array
            The aberrations at the given coordinates.
        """
        return super().eval_basis(coords).sum(0)

    def transmission(
        self: ApertureLayer, coords: Array, pixel_scale: float
    ) -> Array:
        """
        Calculates the transmission of the aperture at the given coordinates.

        Parameters
        ----------
        coords : Array
            The coordinates to calculate the transmission on.
        pixel_scale : float
            The pixel scale of the coordinates.

        Returns
        -------
        transmission : Array
            The transmission of the aperture at the given coordinates.
        """
        if self.transformation is not None:
            coords = self.transformation.apply(coords)
        return self.transmissions(coords, pixel_scale).sum(0)



import jax.numpy as np
from jax import Array

__all__ = [
    "wavenumber",
    "opd2phase",
    "phase2opd",
    "fringe_size",
]


def wavenumber(wavelength: float) -> float:
    """
    Calculates the wavenumber of a given wavelength.

    Parameters
    ----------
    wavelength : Array, metres
        The wavelength to calculate the wavenumber for.

    Returns
    -------
    wavenumber : Array, radians/meter
        The wavenumber of the input wavelength.
    """
    return 2 * np.pi / wavelength


def opd2phase(opd: Array, wavelength: float) -> Array:
    """
    Converts the input Optical Path Difference (opd) in units of meters to phases in
    units of radians for the given wavelength.

    Parameters
    ----------
    opd : Array, metres
        The Optical Path Difference (opd) to be converted into phase.
    wavelength : Array, metres
        The wavelength at which to calculate the phase for.

    Returns
    -------
    phase : Array, radians
        The equivalent phase value for the given opd and wavelength.
    """
    return wavenumber(wavelength) * opd


def phase2opd(phase: Array, wavelength: float) -> Array:
    """
    Converts the input phase in units of radians to the equivalent Optical Path
    Difference (OPD) in metres for the given wavelength.

    Parameters
    ----------
    phase : Array, radians
        The phase to be converted into OPD
    wavelength : Array, metres
        The wavelength at which to calculate the OPD for.

    Returns
    -------
    opd : Array, metres
        The equivalent opd value for the given phase and wavelength.
    """
    return phase / wavenumber(wavelength)


def fringe_size(
    wavelength: float, diameter: float, focal_length: float = None
) -> Array:
    """
    Calculates the linear size of the diffraction fringes.

    Parameters
    ----------
    wavelength : Array, metres
        The wavelength at which to calculate the diffraction fringe for.
    diameter : Array, metres
        The diameter of the aperture.
    focal_length : Array, float = None
        The focal length of the optical system. If none is provided, the fringe
        size is given in units of radians, else it is given in units of metres.

    Returns
    -------
    fringe_size : Array, radians, meters
        The fringe size. Has units of radians of focal length is None, else meters.
    """
    if focal_length is None:
        return wavelength / diameter
    else:
        return wavelength * focal_length / diameter



from __future__ import annotations
from collections import OrderedDict
from abc import abstractmethod
from typing import Union
from zodiax import Base
from jax import Array
import dLux.utils as dlu

from .layers.detector_layers import DetectorLayer
from .psfs import PSF


__all__ = ["BaseDetector", "LayeredDetector"]


class BaseDetector(Base):
    @abstractmethod
    def model(self, psf):  # pragma: no cover
        pass


class LayeredDetector(BaseDetector):
    """
    Applies a series of detector layers to some input psf.

    ??? abstract "UML"
        ![UML](../../assets/uml/LayeredDetector.png)

    Attributes
    ----------
    layers: OrderedDict
        A series of `DetectorLayer` transformations to apply to the input psf.
    """

    layers: OrderedDict

    def __init__(self: LayeredDetector, layers: list[DetectorLayer, tuple]):
        """
        Parameters
        ----------
        layers : list[DetectorLayer, tuple]
            A list of DetectorLayer objects to apply to the input psf. List entries
            can be tuples of (key, layer) to specify a key, else the key is taken as
            the class name of the layer.
        """
        self.layers = dlu.list2dictionary(layers, True, DetectorLayer)
        super().__init__()

    def __getattr__(self: LayeredDetector, key: str) -> object:
        """
        Raises the individual layers via their keys.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the layers dictionary.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the layers
            dictionary.
        """
        if key in self.layers.keys():
            return self.layers[key]
        else:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(type(self), key)
            )

    def model(
        self: LayeredDetector, psf: PSF, return_psf: bool = False
    ) -> Array:
        """
        Applied the detector layers to the input psf.

        Parameters
        ----------
        psf : PSF
            The input psf to be transformed.

        Returns
        -------
        psf : PSF
            The output psf after being transformed by the detector layers.
        """
        for key, layer in self.layers.items():
            psf = layer.apply(psf)
        if return_psf:
            return psf
        return psf.data

    def insert_layer(
        self: LayeredDetector, layer: Union[DetectorLayer, tuple], index: int
    ) -> LayeredDetector:
        """
        Inserts a layer into the layers dictionary at a specified index. This function
        calls the list2dictionary function to ensure all keys remain unique. Note that
        this can result in some keys being modified if they are duplicates. The input
        'layer' can be a tuple of (key, layer) to specify a key, else the key is taken
        as the class name of the layer.

        Parameters
        ----------
        layer : Any
            The layer to be inserted.
        index : int
            The index at which to insert the layer.

        Returns
        -------
        detector : LayeredDetector
            The updated detector.
        """
        return self.set(
            "layers",
            dlu.insert_layer(self.layers, layer, index, DetectorLayer),
        )

    def remove_layer(self: LayeredDetector, key: str) -> LayeredDetector:
        """
        Removes a layer from the layers dictionary, specified by its key.

        Parameters
        ----------
        key : str
            The key of the layer to be removed.

        Returns
        -------
        detector : LayeredDetector
            The updated detector.
        """
        return self.set("layers", dlu.remove_layer(self.layers, key))



import jax.numpy as np
from jax import Array

__all__ = [
    "pad_to",
    "crop_to",
    "resize",
    "downsample",
]

# TODO: Add convolve?


def pad_to(array: Array, npixels: int) -> Array:
    """
    Paraxially zero-pads the input array to the shape (npixels, npixels). Due to the
    paraxial requirement, the input array must be square and even arrays can only
    be padded to even shapes, and odd shaped arrays can only be padded to odd shapes.

    Parameters
    ----------
    array : Array
        The input array to pad.
    npixels : int
        The size to pad to the array to.

    Returns
    -------
    array : Array
        The padded array.
    """
    npixels_in = array.shape[-1]
    if npixels_in % 2 != npixels % 2:
        raise ValueError(
            "Only supports even -> even or odd -> odd padding."
            f"Input array has {npixels_in} pixels, and requested padding to "
            f"{npixels} pixels"
        )
    if npixels < npixels_in:
        raise ValueError(
            "npixels must be larger than the current array, "
            f"npixels_in = {npixels_in} < npixels = {npixels}"
        )

    return np.pad(array, (npixels - npixels_in) // 2)


def crop_to(array: Array, npixels: int) -> Array:
    """
    Paraxially crops the input array to the shape (npixels, npixels). Due to the
    paraxial requirement, the input array must be square and even arrays can only
    be cropped to even shapes, and odd shaped arrays can only be cropped to odd shapes.

    Parameters
    ----------
    array : Array
        The input array to crop.
    npixels : int
        The size to crop the array to.

    Returns
    -------
    array : Array
        The cropped array.
    """
    npixels_in = array.shape[-1]
    if npixels_in % 2 != npixels % 2:
        raise ValueError(
            "Only supports even -> even or odd -> odd cropping."
            f"Input array has {npixels_in} pixels, and requested cropping to "
            f"{npixels} pixels"
        )
    if npixels > npixels_in:
        raise ValueError(
            "npixels must be smaller than the current array, "
            f"npixels_in = {npixels_in} > npixels = {npixels}"
        )

    start, stop = (npixels_in - npixels) // 2, (npixels_in + npixels) // 2
    return array[start:stop, start:stop]


def resize(array: Array, npixels: int) -> Array:
    """
    Resizes the input array to the shape (npixels, npixels), using either a pad or crop
    depending on the input array size. Due to the paraxial requirement, the input array
    must be square and even arrays can only be resized to even shapes, and odd shaped
    arrays can only be resized to odd shapes.

    Parameters
    ----------
    array : Array
        The input array to resize.
    npixels : int
        The size to output the array.

    Returns
    -------
    array : Array
        The resized array.
    """
    npixels_in = array.shape[-1]

    if npixels == npixels_in:
        return array
    elif npixels < npixels_in:
        return crop_to(array, npixels)
    else:
        return pad_to(array, npixels)


def downsample(array: Array, n: int, mean: bool = True) -> Array:
    """
    Downsamples the input array by a factor of n, either by taking the mean or sum
    of the array.

    Parameters
    ----------
    array : Array
        The input array to downsample.
    n : int
        The factor by which to downsample the array.
    mean : bool = True
        Whether to downsample by taking the mean or sum of the array.

    Returns
    -------
    array : Array
        The downsampled array.
    """
    if array.shape[0] != array.shape[1]:
        raise ValueError(
            f"Input array has shape {array.shape}, which is not square"
        )
    if array.shape[0] % n != 0:
        raise ValueError(
            f"Input array has {array.shape[0]} pixels, which is not divisible "
            f"by {n}"
        )

    method = np.mean if mean else np.sum
    size_in = array.shape[0]
    size_out = size_in // n

    # Downsample first dimension
    array = method(array.reshape((size_in * size_out, n)), 1)
    array = array.reshape(size_in, size_out).T

    # Downsample second dimension
    array = method(array.reshape((size_out * size_out, n)), 1)
    array = array.reshape(size_out, size_out).T
    return array



from __future__ import annotations
from jax import Array


from ..wavefronts import Wavefront
from .optical_layers import (
    OpticalLayer,
    TransmissiveLayer,
    BasisLayer,
    AberratedLayer,
)


__all__ = [
    "Optic",
    "BasisOptic",
]


class Optic(TransmissiveLayer, AberratedLayer):
    """
    A basic 'Optic' class, which optionally applies a transmission, OPD and phase to
    the input wavefront, with the option for normalise after.

    ??? abstract "UML"
        ![UML](../../assets/uml/Optic.png)

    Attributes
    ----------
    transmission: Array
        The Array of transmission values to be applied to the input wavefront.
    opd : Array, metres
        The Array of OPD values to be applied to the input wavefront.
    phase : Array, radians
        The Array of phase values to be applied to the input wavefront.
    normalise: bool
        Whether to normalise the wavefront after passing through the optic.
    """

    def __init__(
        self: OpticalLayer,
        transmission: Array = None,
        opd: Array = None,
        phase: Array = None,
        normalise: bool = False,
    ):
        """
        Parameters
        ----------
        transmission: Array = None
            The Array of transmission values to be applied to the input wavefront.
        opd : Array, metres = None
            The Array of OPD values to be applied to the input wavefront.
        phase : Array, radians = None
            The Array of phase values to be applied to the input wavefront.
        normalise: bool = False
            Whether to normalise the wavefront after passing through the optic.
        """
        super().__init__(
            transmission=transmission,
            opd=opd,
            phase=phase,
            normalise=normalise,
        )

        if self.transmission is not None:
            if self.opd is not None:
                if self.transmission.shape != self.opd.shape:
                    raise ValueError(
                        "transmission and opd must have the same shape. Got "
                        f"shapes {self.transmission.shape} and {self.opd.shape}."
                    )
            if self.phase is not None:
                if self.transmission.shape != self.phase.shape:
                    raise ValueError(
                        "transmission and phase must have the same shape. Got "
                        f"shapes {self.transmission.shape} and {self.phase.shape}."
                    )

    def apply(self: OpticalLayer, wavefront: Wavefront) -> Wavefront:
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        wavefront *= self.transmission
        wavefront += self.opd
        wavefront = wavefront.add_phase(self.phase)
        if self.normalise:
            wavefront = wavefront.normalise()
        return wavefront


class BasisOptic(TransmissiveLayer, BasisLayer):
    """
    A basic 'Optic' class, with a aberrations applied through a set of basis vectors
    coefficients. This can be applied either as an opd or phase, using the `as_phase`
    attribute. Also optionally applies a transmission and normalisation.

    ??? abstract "UML"
        ![UML](../../assets/uml/BasisOptic.png)

    Attributes
    ----------
    transmission: Array
        The Array of transmission values to be applied to the input wavefront.
    basis: Array, metres
        Arrays holding the pre-calculated basis vectors.
    coefficients: Array
        The Array of coefficients to be applied to each basis vector.
    as_phase : bool
        Whether to apply the basis as a phase phase or OPD. If True the basis is
        applied as a phase, else it is applied as an OPD.
    normalise : bool
        Whether to normalise the wavefront after passing through the optic.
    """

    def __init__(
        self: OpticalLayer,
        basis,
        transmission=None,
        coefficients=None,
        as_phase=False,
        normalise=False,
    ):
        """
        Parameters
        ----------
        basis: Array, metres
            Arrays holding the pre-calculated basis vectors.
        coefficients: Array = None
            The Array of coefficients to be applied to each basis vector.
        transmission: Array = None
            The Array of transmission values to be applied to the input wavefront.
        as_phase : bool = False
            Whether to apply the basis as a phase phase or OPD. If True the basis is
            applied as a phase, else it is applied as an OPD.
        normalise : bool = False
            Whether to normalise the wavefront after passing through the optic.
        """
        super().__init__(
            transmission=transmission,
            basis=basis,
            coefficients=coefficients,
            as_phase=as_phase,
            normalise=normalise,
        )

    def apply(self: OpticalLayer, wavefront: Wavefront()) -> Wavefront():
        """
        Applies the layer to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The transformed wavefront.
        """
        wavefront *= self.transmission

        if self.as_phase:
            wavefront = wavefront.add_phase(self.eval_basis())
        else:
            wavefront += self.eval_basis()

        if self.normalise:
            wavefront = wavefront.normalise()
        return wavefront



from __future__ import annotations
from zodiax import Base
from jax import Array
import jax.numpy as np
import dLux.utils as dlu


__all__ = ["CoordTransform"]


# Class to be held by dynamic apertures
class CoordTransform(Base):
    """
    A simple class to handle the coordinate transformations applied dynamic
    aperture classes. Transformations are applied in the order:
        1. Translation
        2. Rotation
        3. Compression
        4. Shear

    ??? abstract "UML"
        ![UML](../../assets/uml/CoordTransform.png)

    Attributes
    ----------
    translation: Array
        The (x, y) shift applied to the coords.
    rotation: Array
        The clockwise rotation applied to the coords.
    compression: Array
        The (x, y) compression applied to the coords.
    shear: Array
        The (x, y) shear applied to the coords.
    """

    translation: Array
    rotation: float
    compression: Array
    shear: Array

    def __init__(
        self: CoordTransform,
        translation: Array = None,
        rotation: float = None,
        compression: Array = None,
        shear: Array = None,
    ):
        """
        Parameters
        ----------
        translation: Array
            The (x, y) shift applied to the coords.
        rotation: float, radians
            The clockwise rotation applied to the coords.
        compression: Array
            The (x, y) compression applied to the coords.
        shear: Array
            The (x, y) shear applied to the coords.
        """
        if translation is not None:
            self.translation = np.asarray(translation, dtype=float)
            if self.translation.shape != (2,):
                raise ValueError("center must be have shape (2,).")
        else:
            self.translation = None

        if rotation is not None:
            self.rotation = np.asarray(rotation, dtype=float)
            if self.rotation.shape != ():
                raise ValueError("rotation must have shaoe ().")
        else:
            self.rotation = None

        if compression is not None:
            self.compression = np.asarray(compression, dtype=float)
            if self.compression.shape != (2,):
                raise ValueError("compression must have shape (2,).")
        else:
            self.compression = None

        if shear is not None:
            self.shear = np.asarray(shear, dtype=float)
            if self.shear.shape != (2,):
                raise ValueError("shear must be have shape (2,).")
        else:
            self.shear = None

    def calculate(self: CoordTransform, npix: int, diam: float) -> Array:
        """
        Generate the transformed coords from diameter and npix.

        Parameters
        ----------
        npix : int
            The number of pixels in the output array.
        diam : float
            The diameter of the output array in metres.

        Returns
        -------
        coords : Array
            The transformed coordinates.
        """
        return self.apply(dlu.pixel_coords(npix, diam))

    def apply(self: CoordTransform, coords: Array) -> Array:
        """
        Apply the transformations to the input coordinates.

        Parameters
        ----------
        coords : Array
            The input coordinates to be transformed.

        Returns
        -------
        coords : Array
            The transformed coordinates.
        """
        if self.translation is not None:
            coords = dlu.translate_coords(coords, self.translation)
        if self.shear is not None:
            coords = dlu.shear_coords(coords, self.shear)
        if self.compression is not None:
            coords = dlu.compress_coords(coords, self.compression)
        if self.rotation is not None:
            coords = dlu.rotate_coords(coords, self.rotation)
        return coords



from __future__ import annotations
from abc import abstractmethod
import jax.numpy as np
from zodiax import Base
from jax import vmap, Array

__all__ = ["BaseSpectrum", "Spectrum", "PolySpectrum"]


class BaseSpectrum(Base):
    @abstractmethod
    def normalise(self):  # pragma: no cover
        pass


class SimpleSpectrum(BaseSpectrum):
    """
    Base class for arbitrary spectral parametrisations.

    Attributes
    ----------
    wavelengths : Array, metres
        The array of wavelengths at which the spectrum is defined.
    """

    wavelengths: Array

    def __init__(self: Spectrum, wavelengths: Array):
        """
        Parameters
        ----------
        wavelengths : Array, metres
            The array of wavelengths at which the spectrum is defined.
        """
        self.wavelengths = np.asarray(wavelengths, dtype=float)
        super().__init__()


class Spectrum(SimpleSpectrum):
    """
    A simple spectrum class using wavelengths and weights.


    ??? abstract "UML"
        ![UML](../../assets/uml/Spectrum.png)

    Attributes
    ----------
    wavelengths : Array, metres
        The array of wavelengths at which the spectrum is defined.
    weights : Array
        The relative weights of each wavelength.
    """

    weights: Array

    def __init__(self: Spectrum, wavelengths: Array, weights: Array = None):
        """
        Parameters
        ----------
        wavelengths : Array, metres
            The array of wavelengths at which the spectrum is defined.
        weights : Array = None
            The relative weights of each wavelength. Input weights
            are automatically normalised to a sum of 1.
        """
        super().__init__(wavelengths)
        if weights is None:
            in_shape = self.wavelengths.shape
            weights = np.ones(in_shape) / in_shape[-1]

        weights = np.asarray(weights, dtype=float)
        if weights.ndim == 2:
            self.weights = weights / weights.sum(-1)[:, None]
        else:
            self.weights = weights / weights.sum()

        if self.weights.ndim == 1:
            if self.wavelengths.shape != self.weights.shape:
                raise ValueError(
                    "wavelengths and weights must have the same " "shape."
                )
        else:
            if self.wavelengths.shape != self.weights.shape[-1:]:
                raise ValueError(
                    "wavelengths and weights must have the same " "shape."
                )

    def normalise(self: Spectrum) -> Spectrum:
        """
        Returns a normalised spectrum object, where the weights are normalised to a
        sum of 1.

        Returns
        -------
        spectrum : Spectrum
            The spectrum object with the normalised spectrum.
        """
        if self.weights.ndim == 2:
            weight_sum = self.weights.sum(-1)[:, None]
        else:
            weight_sum = self.weights.sum()
        return self.divide("weights", weight_sum)


class PolySpectrum(SimpleSpectrum):
    """
    Implements a generic polynomial spectrum, such as a linear spectrum.

    This implements a polynomial as follows: f(x) = c0 + c1*x + c2*x^2 + ... + cn*x^n

    ??? abstract "UML"
        ![UML](../../assets/uml/PolySpectrum.png)

    Attributes
    ----------
    wavelengths : Array, metres
        The array of wavelengths at which the spectrum is defined.
    coefficients : Array
        The array of polynomial coefficient values.
    """

    coefficients: Array

    def __init__(self: Spectrum, wavelengths: Array, coefficients: Array):
        """
        Parameters
        ----------
        wavelengths : Array, metres
            The array of wavelengths at which the spectrum is defined.
        coefficients : Array
            The array of polynomial coefficient values.
        """
        super().__init__(wavelengths)
        self.coefficients = np.asarray(coefficients, dtype=float)

        if self.coefficients.ndim != 1:
            raise ValueError("Coefficients must be a 1d array.")

    def _eval_weight(self: Spectrum, wavelength: Array) -> Array:
        """
        Evaluates the polynomial function at the supplied wavelength.

        Parameters
        ----------
        wavelength : Array, metres
            The wavelength at which to evaluate the polynomial function.

        Returns
        -------
        weight : Array
            The relative weight of the supplied wavelength.
        """
        return np.array(
            [
                self.coefficients[i] * wavelength**i
                for i in range(len(self.coefficients))
            ]
        ).sum()

    @property
    def weights(self: Spectrum) -> Array:
        """
        Gets the relative spectral weights by evaluating the polynomial function at the
        internal wavelengths. Output weights are automatically normalised to a sum of
        1.

        Returns
        -------
        weights : Array
            The normalised relative weights of each wavelength.
        """
        weights = vmap(self._eval_weight)(self.wavelengths)
        return weights / weights.sum()

    def normalise(self: Spectrum) -> Spectrum:
        """
        Calculated weights are automatically normalised, so this method simply returns
        an unmodified object.

        Returns
        --------
        spectrum : Spectrum
            The unmodified spectrum object
        """
        return self



# Import as modules
from . import (
    optical_layers,
    optics,
    detector_layers,
    propagators,
    apertures,
    aberrations,
    unified_layers,
)

# Add to __all__
modules = [
    optical_layers,
    optics,
    unified_layers,
    propagators,
    apertures,
    aberrations,
    detector_layers,
]

__all__ = [module.__all__ for module in modules]


from .optical_layers import (
    BaseLayer as BaseLayer,
    TransmissiveLayer as TransmissiveLayer,
    AberratedLayer as AberratedLayer,
    BasisLayer as BasisLayer,
    Tilt as Tilt,
    Normalise as Normalise,
)
from .optics import (
    Optic as Optic,
    BasisOptic as BasisOptic,
)
from .detector_layers import (
    ApplyPixelResponse as ApplyPixelResponse,
    ApplyJitter as ApplyJitter,
    ApplySaturation as ApplySaturation,
    AddConstant as AddConstant,
    Downsample as Downsample,
)
from .unified_layers import (
    Rotate as Rotate,
    Flip as Flip,
    Resize as Resize,
)
from .propagators import (
    MFT as MFT,
    FFT as FFT,
    ShiftedMFT as ShiftedMFT,
    FarFieldFresnel as FarFieldFresnel,
)
from .apertures import (
    CircularAperture as CircularAperture,
    SquareAperture as SquareAperture,
    RectangularAperture as RectangularAperture,
    RegPolyAperture as RegPolyAperture,
    Spider as Spider,
    AberratedAperture as AberratedAperture,
    CompoundAperture as CompoundAperture,
    MultiAperture as MultiAperture,
)
from .aberrations import (
    Zernike as Zernike,
    ZernikeBasis as ZernikeBasis,
)



class Filter(Base):
    """
    NOTE: This class is under development.

    A class for modelling optical filters.

    Attributes
    ----------
    wavelengths : Array
        The wavelengths at which the filter is defined.
    throughput : Array
        The throughput of the filter at the corresponding wavelength.
    filter_name : str
        A string identifier that can be used to initialise specific filters.
    """

    wavelengths: Array
    throughput: Array
    filter_name: str

    def __init__(
        self: Filter,
        wavelengths: Array = None,
        throughput: Array = None,
        filter_name: str = None,
    ) -> Filter:
        """
        Constructor for the Filter class. All inputs are optional and defaults
        to uniform unitary throughput. If filter_name is specified then
        wavelengths and weights must not be specified.

        Parameters
        ----------
        wavelengths : Array = None
            The wavelengths at which the filter is defined.
        throughput : Array = None
            The throughput of the filter at the corresponding wavelength.
        filter_name : str = None
            A string identifier that can be used to initialise specific
            filters. Currently no pre-built filters are implemented.
        """
        # Take the filter name as the priority input
        if filter_name is not None:
            # TODO: Pre load filters
            raise NotImplementedError("You know what this means.")
            pass

            # Check that wavelengths and throughput are not specified
            if wavelengths is not None or throughput is not None:
                raise ValueError(
                    "If filter_name is specified, wavelengths "
                    "and throughput can not be specified."
                )

        # Check that both wavelengths and throughput are specified
        elif (wavelengths is None and throughput is not None) or (
            wavelengths is not None and throughput is None
        ):
            raise ValueError(
                "If either wavelengths or throughput is "
                "specified, then both must be specified."
            )

        # Neither is specified
        elif wavelengths is None and throughput is None:
            self.wavelengths = np.array([0.0, np.inf])
            self.throughput = np.array([1.0, 1.0])
            self.filter_name = "Unitary"

        # Both wavelengths and throughputs are specified
        else:
            self.wavelengths = np.asarray(wavelengths, dtype=float)
            self.throughput = np.asarray(throughput, dtype=float)
            self.filter_name = "Custom"

            # Check bounds
            assert (
                self.wavelengths.ndim == 1 and self.throughput.ndim == 1
            ), "Both wavelengths and throughput must be 1 dimensional."
            assert (
                self.wavelengths.shape == self.throughput.shape
            ), "wavelengths and throughput must have the same length."
            assert (
                np.min(self.wavelengths) >= 0
            ), "wavelengths can not be less than 0."
            assert (self.throughput >= 0).all() and (
                self.throughput <= 1
            ).all(), "throughput must be between 0-1."
            assert np.min(wavelengths) < np.max(
                wavelengths
            ), "wavelengths must be in-order from small to large."

    def get_throughput(self: Filter, sample_wavelengths: Array) -> Array:
        """
        Gets the average throughput of the bandpass defined the differences
        between each sample wavelength, i.e. if sample wavelengths are:
            [10, 20, 30, 40],
        the bandpasses for each sample wavelength will be
            [5-15, 15-25, 25-30, 35-40].
        The throughput is calculated as the average throughput over that
        bandpass.

        Parameters
        ----------
        sample_wavelengths : Array, metres
            The wavelengths at which to sample the filter. Must contain at
            least two values.

        Returns
        -------
        throughputs : Array
            The average throughput for each bandpass defined by
            sample_wavelengths.
        """
        mids = (sample_wavelengths[1:] + sample_wavelengths[:-1]) / 2
        diffs = np.diff(sample_wavelengths)

        start = np.array([sample_wavelengths[0] - diffs[0] / 2])
        end = np.array([sample_wavelengths[-1] + diffs[-1] / 2])
        # min_val = np.array([self.wavelengths.min()])
        # max_val = np.array([self.wavelengths.max()])
        bounds = np.concatenate([start, mids, end])

        # Translate input wavelengths to indexes
        min_wavelength = self.wavelengths.min()
        max_wavelength = self.wavelengths.max()
        num_wavelength = len(self.wavelengths)
        wavelength_range = max_wavelength - min_wavelength
        bnd_indxs = (
            num_wavelength * (bounds - min_wavelength) / wavelength_range
        )
        bnd_indxs = np.clip(bnd_indxs, a_min=0, a_max=len(self.wavelengths))
        bnd_inds = np.round(bnd_indxs, decimals=0).astype(int)

        def nan_div(y, x):
            x_new = np.where(x == 0, 1, x)
            return np.where(x == 0, 0.0, y / x_new)

        def get_tp(start, end, weights, indexes):
            size = end - start
            val = np.where(
                (indexes <= start) | (indexes >= end), 0.0, weights
            ).sum()
            return nan_div(val, size)

        starts = bnd_inds[:-1]
        ends = bnd_inds[1:]
        # dwavelength = self.wavelengths[1] - self.wavelengths[0]
        indexes = np.arange(len(self.wavelengths))

        # weights = self.throughput/self.throughput.sum()
        weights = self.throughput
        out = vmap(get_tp, in_axes=(0, 0, None, None))(
            starts, ends, weights, indexes
        )
        return out

    def model(self: Filter, optics: Optics, **kwargs):
        """
        A base level modelling function designed to robustly handle the
        different combinations of inputs. Models the sources through the
        instrument optics and detector. Users must provide optics and some form
        of source, either via a scene, sources or single source input, but not
        multiple.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector to use with the observation. Defaults to the
            internally stored value.
        scene : Scene = None
            The scene to observe. Defaults to the internally stored value.
        sources : Union[dict, list, tuple) = None
            The sources to observe.
        source : dLux.sources.Source = None
            The source to observe.
        normalise_sources : bool = True
            Whether to normalise the sources before modelling. Default is True.
        flatten : bool = False
            Whether the output image should be flattened. Default is False.
        return_tree : bool = False
            Whether to return a Pytree like object with matching tree structure
            as the input scene/sources/source. Default is False.

        Returns
        -------
        image : Array, Pytree
            The image of the scene modelled through the optics with detector
            and filter effects applied if they are supplied. Returns either as
            a single array (if return_tree is false), or a pytree like object
            with matching tree structure as the input scene/sources/source.
        """
        return model(optics, filter=self, **kwargs)



from __future__ import annotations
from abc import abstractmethod
from zodiax import Base
from jax import Array, vmap
import jax.numpy as np
from typing import Union

from .optical_systems import BaseOpticalSystem as OpticalSystem
from .detectors import BaseDetector as Detector
from .sources import BaseSource as Source, Scene
from .psfs import PSF


__all__ = ["Instrument", "Telescope", "Dither"]


class Instrument(Base):
    @abstractmethod
    def model(self):  # pragma: no cover
        pass


class Telescope(Instrument):
    """
    Class that represents a telescope instrument, holding an optical system, a source
    object and (optionally) a detector object, automating the process of modelling
    all three in conjunction.

    To generate more complex instruments or a set of observations, the `Telescope`
    class can be inherited and modified to suit the needs of the user.

    Attributes
    ----------
    optics : OpticalSystem
        An `OpticalSystem` object that defines the optical transformations of the
        instrument.
    source : Source
        A `Source` or `Scene` to objects to model through the instrument.
    detector : Detector
        A `Detector` object that defines the detector transformations of the
        instrument.
    """

    optics: OpticalSystem
    source: Source
    detector: Detector

    def __init__(
        self: Telescope,
        optics: OpticalSystem,
        source: Union[list, Source],
        detector: Detector = None,
    ):
        """
        Parameters
        ----------
        optics : OpticalSystem
            An `OpticalSystem` object that defines the optical transformations of the
            instrument.
        source : Source
            A `Source` or `Scene` to objects to model through the instrument. Can be
            either a single `Source` object, or a list of `Source` objects which is
            then converted to a `Scene` object. The list entries can also be a tuple of
            (key, source)  in order to specify a key for the source in the scene.
        detector : Detector = None
            A `Detector` object that defines the detector transformations of the
            instrument.
        """
        # Optics
        if not isinstance(optics, OpticalSystem):
            raise TypeError("optics must be an Optics object.")
        self.optics = optics

        # Sources
        if isinstance(source, Source):
            self.source = source
        elif isinstance(source, tuple):
            # If its a (key, source) tuple, we ignore the key
            self.source = source[1]
        else:
            self.source = Scene(source)

        # Detector
        if not isinstance(detector, Detector) and detector is not None:
            raise TypeError(
                "detector must be an Detector object. "
                f"Got type {type(detector)}"
            )
        self.detector = detector

    def __getattr__(self: Telescope, key: str) -> object:
        """
        Raises the attributes from the optics, source and detector to the top level of
        the class.

        Parameters
        ----------
        key : str
            The key of the item to be searched for.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the sub-dictionaries.
        """
        for attribute in self.__dict__.values():
            if hasattr(attribute, key):
                return getattr(attribute, key)
        raise AttributeError(
            f"{self.__class__.__name__} has no attribute " f"{key}."
        )

    def model(self: Telescope, return_psf: bool = False) -> Array:
        """
        Models the source objects through the optical system and detector.

        Parameters
        ----------
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, PSF
            if `return_psf` is False, the psf Array is returned.
            If `return_psf` is True, the PSF object is returned.

        """
        # Model optics: return_psf=True for more efficient source calculations
        psfs = self.optics.model(self.source, return_psf=True)

        # Array based output
        psf = psfs.data.sum(tuple(range(psfs.ndim)))
        pixel_scale = psfs.pixel_scale.mean()

        # Pass through detector transformations if it exists
        psf_obj = PSF(psf, pixel_scale)
        if self.detector is not None:
            return self.detector.model(psf_obj, return_psf=return_psf)

        # Return psf
        if return_psf:
            return psf_obj
        return psf_obj.data


class Dither(Telescope):
    """
    Simple extension of the `Telescope` class that applies a series of dithers to the
    source positions before modelling the instrument. Serves both as a demonstration
    of how to extend the `Telescope` class and as a useful tool for modelling
    dithered observations.

    Attributes
    ----------
    optics : OpticalSystem
        An `OpticalSystem` object that defines the optical transformations of the
        instrument.
    source : Source
        A `Source` or `Scene` to objects to model through the instrument.
    detector : Detector
        A `Detector` object that defines the detector transformations of the
        instrument.
    dithers : Array, radians
        The array of dithers to apply to the source positions. The shape of the
        array should be (ndithers, 2).
    """

    dithers: Array

    def __init__(
        self: Telescope,
        dithers: Array,
        optics: OpticalSystem,
        source: Union[list, Source],
        detector: Detector = None,
    ):
        """
        Parameters
        ----------
        dithers : Array, radians
            The array of dithers to apply to the source positions. The shape of the
            array should be (ndithers, 2).
        optics : OpticalSystem
            An `OpticalSystem` object that defines the optical transformations of the
            instrument.
        source : Source
            A `Source` or `Scene` to objects to model through the instrument. Can be
            either a single `Source` object, or a list of `Source` objects which is
            then converted to a `Scene` object. The list entries can also be a tuple of
            (key, source)  in order to specify a key for the source in the scene.
        detector : Detector = None
            A `Detector` object that defines the detector transformations of the
            instrument.
        """
        self.dithers = np.asarray(dithers, float)
        if self.dithers.ndim != 2 or self.dithers.shape[1] != 2:
            raise ValueError("dithers must be an array of shape (ndithers, 2)")
        super().__init__(optics=optics, source=source, detector=detector)

    def model(self: Telescope, return_psf: bool = False) -> Array:
        """
        Models the source objects through the optical system and detector, while also
        applying the dithers to the source positions.

        Parameters
        ----------
        return_psf : bool = False
            Should the PSF object be returned instead of the psf Array?

        Returns
        -------
        object : Array, PSF
            if `return_psf` is False, the psf Array is returned.
            If `return_psf` is True, the PSF object is returned.
        """

        def dither_and_model(dither, instrument):
            instrument = instrument.add("source.position", dither)
            return super(type(instrument), instrument).model(return_psf)

        return vmap(dither_and_model, (0, None))(self.dithers, self)



from __future__ import annotations
import jax.numpy as np
from jax.scipy.signal import convolve
from jax import Array
from zodiax import Base
import dLux.utils as dlu


__all__ = ["PSF"]


class PSF(Base):
    """
    A simple class that holds the state of some PSF as it it transformed by detector
    layers.

    Attributes
    ----------
    data : Array
        The psf as it is transformed by the detector.
    pixel_scale : Array
        The pixel scale of the psf.
    """

    data: Array
    pixel_scale: Array

    def __init__(self: PSF, data: Array, pixel_scale: Array):
        """
        Parameters
        ----------
        data : Array
            The psf to be transformed by the detector.
        pixel_scale : Array
            The pixel scale of the psf.
        """
        self.data = np.asarray(data, dtype=float)
        self.pixel_scale = np.asarray(pixel_scale, dtype=float)

    @property
    def npixels(self: PSF) -> int:
        """
        Returns the side length of the arrays currently representing the psf.

        Returns
        -------
        npixels : int
            The number of pixels that represent the `PSF`.
        """
        return self.data.shape[-1]

    @property
    def ndim(self: PSF) -> int:
        """
        Returns the number of 'dimensions' of the psf. This is used to track the
        vectorised version of the psf returned from vmapping.

        Returns
        -------
        ndim : int
            The 'dimensionality' of dimensions of the psf.
        """
        return self.pixel_scale.ndim

    def downsample(self: PSF, n: int) -> PSF:
        """
        Downsamples the psf by a factor of n. This is done by summing the psf pixels in
        n x n blocks.

        Parameters
        ----------
        n : int
            The factor by which to downsample the psf.

        Returns
        -------
        psf : PSF
            The downsampled psf.
        """
        data = dlu.downsample(self.data, n, mean=False)
        pixel_scale = self.pixel_scale * n
        return self.set(["data", "pixel_scale"], [data, pixel_scale])

    def convolve(self: PSF, other: Array, method: str = "auto") -> PSF:
        """
        Convolves the psf with some input array.

        Parameters
        ----------
        other : Array
            The psf to convolve with.
        method : str = "auto"
            The method to use for the convolution. Can be "auto", "direct",
            or "fft". Is "auto" by default, which calls "direct".

        Returns
        -------
        psf : PSF
            The convolved psf.
        """
        return self.set(
            "data", convolve(self.data, other, mode="same", method=method)
        )

    def rotate(self: PSF, angle: float, order: int = 1) -> PSF:
        """
        Rotates the psf by a given angle via interpolation.

        Parameters
        ----------
        angle : float
            The angle by which to rotate the psf.
        order : int = 1
            The order of the interpolation method to use.

        Returns
        -------
        psf : PSF
            The rotated psf.
        """
        return self.set("data", dlu.rotate(self.data, angle, order=order))

    def resize(self: PSF, npixels: int) -> PSF:
        """
        Resizes the psf via a zero-padding or cropping operation.

        Parameters
        ----------
        npixels : int
            The size to resize the psf to.

        Returns
        -------
        psf : PSF
            The resized psf.
        """
        return self.set("data", dlu.resize(self.data, npixels))

    def flip(self: PSF, axis: tuple) -> PSF:
        """
        Flips the psf along the specified axes. Note we use 'ij' indexing, so axis 0 is
        the y-axis and axis 1 is the x-axis.

        Parameters
        ----------
        axis : tuple
            The axes along which to flip the PSF.

        Returns
        -------
        psf : PSF
            The new flipped PSF.
        """
        return self.set("data", np.flip(self.data, axis))

    def __mul__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the multiplication operator. This allows for the
        multiplication of the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to multiply the psf by.

        Returns
        -------
        psf : PSF
            The multiplied psf.
        """
        return self.multiply("data", other)

    def __imul__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the inplace multiplication operator. This allows for the
        inplace multiplication of the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to multiply the psf by.

        Returns
        -------
        psf : PSF
            The multiplied psf.
        """
        return self.__mul__(other)

    def __add__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the addition operator. This allows for the addition of the psf
        by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to add to the psf.

        Returns
        -------
        psf : PSF
            The added psf.
        """
        return self.add("data", other)

    def __iadd__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the inplace addition operator. This allows for the inplace
        addition of the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to add to the psf.

        Returns
        -------
        psf : PSF
            The added psf.
        """
        return self.__add__(other)

    def __sub__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the subtraction operator. This allows for the subtraction of
        the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to subtract from the psf.

        Returns
        -------
        psf : PSF
            The subtracted psf.
        """
        return self.add("data", -other)

    def __isub__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the inplace subtraction operator. This allows for the inplace
        subtraction of the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to subtract from the psf.

        Returns
        -------
        psf : PSF
            The subtracted psf.
        """
        return self.__sub__(other)

    def __truediv__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the division operator. This allows for the division of the psf
        by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to divide the psf by.

        Returns
        -------
        psf : PSF
            The divided psf.
        """
        return self.divide("data", other)

    def __itruediv__(self: PSF, other: Array) -> PSF:
        """
        Magic method for the inplace division operator. This allows for the inplace
        division of the psf by a scalar or another psf.

        Parameters
        ----------
        other : Array
            The scalar or psf to divide the psf by.

        Returns
        -------
        psf : PSF
            The divided psf.
        """
        return self.__truediv__(other)



from __future__ import annotations
import jax.numpy as np
from typing import Any
from jax import vmap, Array
from zodiax import Base
import dLux.utils as dlu
import dLux

# Optical layers require Wavefront to init, so we alias it here to avoid MRO issues
OpticalLayer = lambda: dLux.layers.optical_layers.OpticalLayer


__all__ = ["Wavefront"]


class Wavefront(Base):
    """
    A simple class to hold the state of some wavefront as it is transformed and
    propagated throughout an optical system. All wavefronts assume square arrays.

    Attributes
    ----------
    wavelength : float, meters
        The wavelength of the `Wavefront`.
    amplitude : Array, power
        The electric field amplitude of the `Wavefront`.
    phase : Array, radians
        The electric field phase of the `Wavefront`.
    pixel_scale : float, meters/pixel or radians/pixel
        The pixel scale of the phase and amplitude arrays. If `units='Cartesian'` then
        the pixel scale is in meters/pixel, else if `units='Angular'` then the pixel
        scale is in radians/pixel.
    plane : str
        The current plane type of wavefront, can be 'Pupil', 'Focal' or 'Intermediate'.
    units : str
        The current units of the wavefront, can be 'Cartesian' or 'Angular'.
    """

    wavelength: float
    pixel_scale: float
    amplitude: Array
    phase: Array
    plane: str
    units: str

    def __init__(
        self: Wavefront, npixels: int, diameter: float, wavelength: float
    ):
        """
        Parameters
        ----------
        npixels : int
            The number of pixels that represent the `Wavefront`.
        diameter : float, meters
            The total diameter of the `Wavefront`.
        wavelength : float, meters
            The wavelength of the `Wavefront`.
        """
        self.wavelength = np.asarray(wavelength, float)
        self.pixel_scale = np.asarray(diameter / npixels, float)
        self.amplitude = (
            np.ones((npixels, npixels), dtype=float) / npixels**2
        )
        self.phase = np.zeros((npixels, npixels), dtype=float)

        # Always initialised in Pupil plane with Cartesian Coords
        self.plane = "Pupil"
        self.units = "Cartesian"

    ####################
    # Getter Functions #
    ####################
    @property
    def diameter(self: Wavefront) -> Array:
        """
        Returns the current wavefront diameter calculated using the pixel scale and
        number of pixels.

        Returns
        -------
        diameter : Array, meters or radians
            The current diameter of the wavefront.
        """
        return self.npixels * self.pixel_scale

    @property
    def npixels(self: Wavefront) -> int:
        """
        Returns the side length of the arrays currently representing the wavefront.
        Taken from the last axis of the amplitude array.

        Returns
        -------
        pixels : int
            The number of pixels that represent the `Wavefront`.
        """
        return self.amplitude.shape[-1]

    @property
    def real(self: Wavefront) -> Array:
        """
        Returns the real component of the `Wavefront`.

        Returns
        -------
        wavefront : Array
            The real component of the `Wavefront` phasor.
        """
        return self.amplitude * np.cos(self.phase)

    @property
    def imaginary(self: Wavefront) -> Array:
        """
        Returns the imaginary component of the `Wavefront`.

        Returns
        -------
        wavefront : Array
            The imaginary component of the `Wavefront` phasor.
        """
        return self.amplitude * np.sin(self.phase)

    @property
    def phasor(self: Wavefront) -> Array:
        """
        The electric field phasor described by this Wavefront in complex form.

        Returns
        -------
        field : Array
            The electric field phasor of the wavefront.
        """
        return self.amplitude * np.exp(1j * self.phase)

    @property
    def psf(self: Wavefront) -> Array:
        """
        Calculates the Point Spread Function (PSF), i.e. the squared modulus
        of the complex wavefront.

        Returns
        -------
        psf : Array
            The PSF of the wavefront.
        """
        return self.amplitude**2

    @property
    def coordinates(self: Wavefront) -> Array:
        """
        Returns the physical positions of the wavefront pixels in meters.

        Returns
        -------
        coordinates : Array
            The coordinates of the centers of each pixel representing the
            wavefront.
        """
        return dlu.pixel_coords(self.npixels, self.diameter)

    @property
    def wavenumber(self: Wavefront) -> Array:
        """
        Returns the wavenumber of the wavefront (2 * pi / wavelength).

        Returns
        -------
        wavenumber : Array, 1/meters
            The wavenumber of the wavefront.
        """
        return 2 * np.pi / self.wavelength

    @property
    def fringe_size(self: Wavefront) -> Array:
        """
        Returns the size of the fringes in angular units.

        TODO Units check from focal plane
        Returns
        -------
        fringe_size : Array, radians
            The size of the linear diffraction fringe of the wavefront.
        """
        return self.wavelength / self.diameter

    @property
    def ndim(self: Wavefront) -> int:
        """
        Returns the number of 'dimensions' of the wavefront. This is used to track the
        vectorised version of the wavefront returned from vmapping.

        Returns
        -------
        ndim : int
            The 'dimensionality' of dimensions of the wavefront.
        """
        return self.pixel_scale.ndim

    #################
    # Magic Methods #
    #################
    def __add__(self: Wavefront, other: Any) -> Wavefront:
        """
        Adds the input 'other' to the wavefront. If the input is a numeric type, it is
        treated as an OPD, else if it is an optical layer, it will be applied to the
        wavefront.

        Parameters
        ----------
        other : Any
            The input to add to the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The output wavefront.
        """
        # None Type
        if other is None:
            return self

        # Some Optical Layer
        if isinstance(other, OpticalLayer()):
            return other.apply(self)

        # Array based inputs - Defaults to OPD
        if isinstance(other, (Array, float, int)):
            return self.add_opd(other)

        # Other
        else:
            raise TypeError(
                "Can only add an array or OpticalLayer to "
                f"Wavefront. Got: {type(other)}."
            )

    def __iadd__(self: Wavefront, other: Any) -> Wavefront:
        """
        Provides the += operator for the wavefront, calling the __add__ method.

        Parameters
        ----------
        other : Any
            The input to add to the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The output wavefront.
        """
        return self.__add__(other)

    def __mul__(self: Wavefront, other: Any) -> Wavefront:
        """
        Multiplies the input 'other' to the wavefront. If the input is a numeric type,
        it is treated as an array of transmission values and is multiplied by the
        wavefront amplitude, unless it is a complex number, in which case it will be
        multiplied with the wavefront phasor. If it is an optical layer, it will be
        applied to the wavefront.

        Parameters
        ----------
        other : Any
            The input to multiply with the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The output wavefront.
        """
        # None Type, return None
        if other is None:
            return self

        # Some Optical Layer, apply it
        if isinstance(other, OpticalLayer()):
            return other.apply(self)

        # Array based inputs
        if isinstance(other, (Array, float, int)):
            # Complex array - Multiply the phasors
            if isinstance(other, Array) and other.dtype.kind == "c":
                phasor = self.phasor * other
                return self.set(
                    ["amplitude", "phase"], [np.abs(phasor), np.angle(phasor)]
                )

            # Scalar array - Multiply amplitude
            else:
                return self.multiply("amplitude", other)

        # Other
        else:
            raise TypeError(
                "Can only multiply Wavefront by array or "
                f"OpticalLayer. Got: {type(other)}."
            )

    def __imul__(self: Wavefront, other: Any) -> Wavefront:
        """
        Provides the *= operator for the wavefront, calling the __mul__ method.

        Parameters
        ----------
        other : Any
            The input to multiply with the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The output wavefront.
        """
        return self.__mul__(other)

    ###################
    # Adder Functions #
    ###################
    def add_opd(self: Wavefront, opd: Array) -> Wavefront:
        """
        Adds an optical path difference (OPD) to the wavefront.

        Parameters
        ----------
        opd : Array, meters
            The opd to add to the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The new wavefront with the phases updated according to the supplied opd.
        """
        return self.add("phase", self.wavenumber * opd)

    def add_phase(self: Wavefront, phase: Array) -> Wavefront:
        """
        Adds a phase to the wavefront.

        Parameters
        ----------
        phase : Array, radians
            The phase to be added to the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The new wavefront with updated phases.
        """
        # Add this extra None check to allow PhaseOptics to have a None phase
        # and still be able to be 'added' to it, making this the phase
        # equivalent of `wf += opd` -> `wf = wf.add_phase(phase)`
        if phase is not None:
            return self.add("phase", phase)
        return self

    ###################
    # Other Functions #
    ###################
    def tilt(self: Wavefront, angles: Array) -> Wavefront:
        """
        Tilts the wavefront by the (x, y) angles.

        Parameters
        ----------
        angles : Array, radians
            The (x, y) angles by which to tilt the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The tilted wavefront.
        """
        if not isinstance(angles, Array) or angles.shape != (2,):
            raise ValueError("angles must be an array of shape (2,).")
        opd = -(angles[:, None, None] * self.coordinates).sum(0)
        return self.add_opd(opd)

    def normalise(self: Wavefront) -> Wavefront:
        """
        Normalises the total power of the wavefront to 1.

        Returns
        -------
        wavefront : Wavefront
            The normalised wavefront.
        """
        return self.divide("amplitude", np.linalg.norm(self.amplitude))

    def _to_field(self: Wavefront, complex: bool = False) -> Array:
        """
        Returns the wavefront in either (amplitude, phase) or (real, imaginary) form.

        Parameters
        ----------
        complex : bool = False
            Whether to return the wavefront in (real, imaginary) form.

        Returns
        -------
        field : Array
            The wavefront in either (amplitude, phase) or (real, imaginary) form.
        """
        if complex:
            return np.array([self.real, self.imaginary])
        return np.array([self.amplitude, self.phase])

    def _to_amplitude_phase(self: Wavefront, field: Array) -> Array:
        """
        Transforms the input field in (real, imaginary) to (amplitude, phase) form.

        Parameters
        ----------
        field : Array
            The wavefront field in (real, imaginary) form.

        Returns
        -------
        field : Array
            The wavefront field in (amplitude, phase) form.
        """
        amplitude = np.hypot(field[0], field[1])
        phase = np.arctan2(field[1], field[0])
        return np.array([amplitude, phase])

    def flip(self: Wavefront, axis: tuple) -> Wavefront:
        """
        Flips the wavefront along the specified axes. Note we use 'ij' indexing, so
        axis 0 is the y-axis and axis 1 is the x-axis.

        Parameters
        ----------
        axis : tuple
            The axes along which to flip the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The new flipped wavefront.
        """
        field = self._to_field()
        flipper = vmap(np.flip, (0, None))
        amplitude, phase = flipper(field, axis)
        return self.set(["amplitude", "phase"], [amplitude, phase])

    def scale_to(
        self: Wavefront,
        npixels: int,
        pixel_scale: Array,
        complex: bool = False,
    ) -> Wavefront:
        """
        Interpolated the wavefront to a given npixels and pixel_scale. Can be done on
        the real and imaginary components by passing in complex=True.

        Parameters
        ----------
        npixels : int
            The number of pixels  to interpolate to.
        pixel_scale: Array
            The pixel scale to interpolate to.
        complex : bool = False
            Whether to rotate the real and imaginary representation of the wavefront as
            opposed to the amplitude and phase representation.

        Returns
        -------
        wavefront : Wavefront
            The new interpolated wavefront.
        """
        # Get field in either (amplitude, phase) or (real, imaginary)
        field = self._to_field(complex=complex)

        # Scale the field
        scale_fn = vmap(dlu.scale, (0, None, None))
        field = scale_fn(field, npixels, pixel_scale / self.pixel_scale)

        # Cast back to (amplitude, phase) if needed
        if complex:
            field = self._to_amplitude_phase(field)

        # Return new wavefront
        return self.set(
            ["amplitude", "phase", "pixel_scale"],
            [field[0], field[1], pixel_scale],
        )

    def rotate(
        self: Wavefront, angle: Array, order: int = 1, complex: bool = False
    ) -> Wavefront:
        """
        Rotates the wavefront by a given angle via interpolation. Can be done on the
        real and imaginary components by passing in complex=True.

        Parameters
        ----------
        angle : Array, radians
            The angle by which to rotate the wavefront in a clockwise
            direction.
        order : int = 1
            The interpolation order to use.
        complex : bool = False
            Whether to rotate the real and imaginary representation of the wavefront as
            opposed to the amplitude and phase representation.

        Returns
        -------
        wavefront : Wavefront
            The new wavefront rotated by angle in the clockwise direction.
        """
        # Get field in either (amplitude, phase) or (real, imaginary)
        field = self._to_field(complex=complex)

        # Rotate the field
        rotator = vmap(dlu.rotate, (0, None, None))
        field = rotator(field, angle, order)

        # Cast back to (amplitude, phase) if needed
        if complex:
            field = self._to_amplitude_phase(field)

        # Return new wavefront
        return self.set(["amplitude", "phase"], [field[0], field[1]])

    def resize(self: Wavefront, npixels: int) -> Wavefront:
        """
        Resizes the wavefront via a zero-padding or cropping operation.

        Parameters
        ----------
        npixels : int
            The size to resize the wavefront to.

        Returns
        -------
        wavefront : Wavefront
            The resized wavefront.
        """
        field = self._to_field()
        amplitude, phase = vmap(dlu.resize, (0, None))(field, npixels)
        return self.set(["amplitude", "phase"], [amplitude, phase])

    #########################
    # Propagation Functions #
    #########################
    def _prep_prop(self: Wavefront, focal_length) -> tuple:
        """
        Determines the propagation direction, output plane and output units.

        Parameters
        ----------
        focal_length : Union[float, None]
            The focal length of the propagation.

        Returns
        -------
        inverse : bool
            Whether the propagation is inverse or not.
        plane : str
            The output plane of the propagation.
        units : str
            The output units of the propagation.
        """
        # Determine propagation direction, output plane and output units
        if self.plane == "Pupil":
            inverse = False
            plane = "Focal"
            if focal_length is None:
                units = "Angular"
            else:
                units = "Cartesian"
        else:
            if focal_length is not None and self.units == "Angular":
                raise ValueError(
                    "focal_length can not be specific when"
                    "propagating from a Focal plane with angular units."
                )
            inverse = True
            plane = "Pupil"
            units = "Cartesian"

        return inverse, plane, units

    def propagate_FFT(
        self: Wavefront,
        focal_length: float = None,
        pad: int = 2,
    ) -> Wavefront:
        """
        Propagates the wavefront by performing a Fast Fourier Transform.

        Parameters
        ----------
        focal_length : float = None
            The focal length of the propagation. If None, the output pixel scale has
            units of radians, else meters.
        pad : int = 2
            The padding factory to apply to the input wavefront before the FFT.

        Returns
        -------
        wavefront : Wavefront
            The propagated wavefront.
        """
        inverse, plane, units = self._prep_prop(focal_length)

        # Calculate
        phasor, pixel_scale = dlu.FFT(
            self.phasor,
            self.wavelength,
            self.pixel_scale,
            focal_length,
            pad,
            inverse,
        )

        # Return new wavefront
        return self.set(
            ["amplitude", "phase", "pixel_scale", "plane", "units"],
            [np.abs(phasor), np.angle(phasor), pixel_scale, plane, units],
        )

    # TODO: Class method this?
    def _MFT(
        self: Wavefront,
        phasor: Array,
        wavelength: float,
        pixel_scale: float,
        *args: tuple,
    ) -> Array:
        """
        Simple alias for the MFT function to allow for vectorisation over phasors,
        wavelengths, pixel_scales, etc.

        Parameters
        ----------
        phasor : Array
            The phasor to propagate.
        wavelength : float
            The wavelength of the wavefront.
        pixel_scale : float
            The pixel scale of the wavefront.
        args : tuple
            The propagation arguments.

        Returns
        -------
        phasor : Array
            The propagated phasor.
        """
        return dlu.MFT(phasor, wavelength, pixel_scale, *args)

    def propagate(
        self: Wavefront,
        npixels: int,
        pixel_scale: float,
        focal_length: float = None,
        shift: Array = np.zeros(2),
        pixel: bool = True,
    ) -> Wavefront:
        """
        Propagates the wavefront by performing an MFT, allowing for the output pixel
        scale and npixels to be specified.

        Parameters
        ----------
        npixels : int
            The number of pixels in the output plane.
        pixel_scale : float, meters/pixel or radians/pixel
            The pixel scale of the output plane.
        focal_length : float = None
            The focal length of the propagation. If None, the propagation is angular
            and pixel_scale_out is taken in as radians/pixel, else meters/pixel.
        shift : Array = np.zeros(2)
            The shift in the center of the output plane.
        pixel : bool = True
            Should the shift be taken in units of pixels, or pixel scale.

        Returns
        -------
        wavefront : Wavefront
            The propagated wavefront.
        """
        inverse, plane, units = self._prep_prop(focal_length)

        # Enforce array so output can be vectorised by vmap
        pixel_scale = np.asarray(pixel_scale, float)

        # Calculate
        # Using a self._MFT here allows for broadband wavefronts to define
        # vectorised propagation fn over phasors, wavels, px_scales, etc.
        # It also makes the code muuuuch nicer to read
        args = (npixels, pixel_scale, focal_length, shift, pixel, inverse)
        phasor = self._MFT(
            self.phasor, self.wavelength, self.pixel_scale, *args
        )

        # Update
        return self.set(
            ["amplitude", "phase", "pixel_scale", "plane", "units"],
            [np.abs(phasor), np.angle(phasor), pixel_scale, plane, units],
        )

    # # TODO: Class method this?
    # def _fresnel(self, phasor, wavelength, pixel_scale, focal_shift, *args):
    #     return dlu.fresnel_MFT(phasor, wavelength, pixel_scale, *args)

    def propagate_fresnel(
        self: Wavefront,
        npixels: int,
        pixel_scale: float,
        focal_length: float,
        focal_shift: float = 0.0,
        shift: Array = np.zeros(2),
        pixel: bool = True,
    ) -> Wavefront:
        """
        Propagates the phasor using a Far-Field Fresnel propagation. This allows for
        psfs to be better modelled a few wavelengths from the focal plane.

        Parameters
        ----------
        npixels : int
            The number of pixels in the output plane.
        pixel_scale : float, meters/pixel or radians/pixel
            The pixel scale of the output plane.
        focal_length : float
            The focal length of the propagation.
        focal_shift: float, meters = 0.
            The shift from focus to propagate to.
        shift : Array = np.zeros(2)
            The shift in the center of the output plane.
        pixel : bool = True
            Should the shift be taken in units of pixels, or pixel scale.

        Returns
        -------
        wavefront : Wavefront
            The propagated wavefront.
        """
        # TODO: Try inverse propagation to see if it works, it probably will
        if self.plane == "Pupil":
            inverse = False
        else:
            inverse = True
        plane = "Intermediate"
        units = "Cartesian"

        # We can't fresnel from a focal plane
        if self.plane != "Pupil":
            raise ValueError(
                "Can only do an fresnel propagation from a Pupil plane, "
                f"current plane is {self.plane}."
            )

        # Calculate
        phasor = dlu.fresnel_MFT(
            self.phasor,
            self.wavelength,
            self.pixel_scale,
            npixels,
            pixel_scale,
            focal_length,
            focal_shift,
            shift,
            pixel,
            inverse,
        )

        # Update
        return self.set(
            ["amplitude", "phase", "pixel_scale", "plane", "units"],
            [np.abs(phasor), np.angle(phasor), pixel_scale, plane, units],
        )



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import LayeredDetector, PSF
from dLux.layers import Downsample


psf = PSF(np.ones((16, 16)), 1 / 16)


def test_layered_detector():
    det = LayeredDetector([Downsample(4)])

    # Test getattr
    det.Downsample
    with pytest.raises(AttributeError):
        det.not_an_attr

    # Test model
    assert isinstance(det.model(psf), np.ndarray)
    assert isinstance(det.model(psf, return_psf=True), PSF)

    # Test insert and remove layer
    det.insert_layer(Downsample(4), 1)
    det.remove_layer("Downsample")



import dLux
import typing
import jax
import jax.numpy as np
import equinox as eqx

Scalar = typing.NewType("Scalar", np.ndarray)  # 0d
Vector = typing.NewType("Vector", np.ndarray)  # 1d
Array = typing.NewType("Array", np.ndarray)  # 2d +

# Base ∂Lux Objects
Wavefront = typing.NewType("Wavefront", dLux.Wavefront)
Propagator = typing.NewType("Propagator", dLux.Propagator)
PlaneType = typing.NewType("PlaneType", dLux.PlaneType)
Layer = typing.NewType("Layers", object)


class GaussianWavefront(dLux.Wavefront):
    """
    Expresses the behaviour and state of a wavefront propagating in
    an optical system under the fresnel assumptions. This
    implementation is based on the same class from the `poppy`
    library [poppy](https://github.com/spacetelescope/poppy/fresnel.py)
    and Chapter 3 from _Applied Optics and Optical Engineering_
    by Lawrence G. N.

    Approximates the wavefront as a Gaussian Beam parameterised by the
    radius of the beam, the phase radius, the phase factor and the
    Rayleigh distance. Propagation is based on two different regimes
    for a total of four different operations.

    Attributes
    ----------
    position : float, metres
        The position of the wavefront in the optical system.

    beam_radius : float, metres
        The *initial* radius of the beam.


    waist_position : float, metres
        The position of the beam waist along the optical axis.
    spherical : bool
        A convenience tracker for knowing if the wavefront is
        currently spherical.
    rayleigh_factor : float
        Used to determine the range over which the wavefront remains
        planar.
    focal_length : float, metres
        Used for the conversion between angular and physical units.
    """

    angular: bool
    spherical: bool
    waist_radius: Scalar
    position: Scalar
    waist_position: Scalar
    rayleigh_factor: Scalar
    focal_length: Scalar

    def __init__(
        self: Wavefront,
        wavelength: Scalar,
        offset: Vector,
        pixel_scale: Scalar,
        plane_type: PlaneType,
        amplitude: Array,
        phase: Array,
        waist_radius: Scalar,
        rayleigh_factor: Scalar = 2.0,
    ) -> Wavefront:
        """
        Creates a wavefront with an empty amplitude and phase
        arrays but of a given wavelength and phase offset.
        Assumes that the beam starts at the waist following from
        the `poppy` convention.

        Parameters
        ----------
        beam_radius : float, metres
            Radius of the beam at the initial optical plane.
        wavelength : float, metres
            The wavelength of the `Wavefront`.
        offset : Array, radians
            The (x, y) angular offset of the `Wavefront` from
            the optical axis.
        rayleigh_factor : float
            A multiplicative factor determining the threshold for
            considering the wavefront spherical.
        """
        super().__init__(
            wavelength, offset, pixel_scale, plane_type, amplitude, phase
        )
        self.waist_radius = np.asarray(waist_radius).astype(float)
        self.position = np.asarray(0.0, dtype=float)
        self.waist_position = np.asarray(0.0, dtype=float)
        self.rayleigh_factor = np.asarray(rayleigh_factor, dtype=float)
        self.focal_length = np.inf
        self.angular = np.asarray(False, dtype=bool)
        self.spherical = np.asarray(False, dtype=bool)

    # Getters #
    def get_position(self: Wavefront) -> float:
        """
        Accessor for the position of the wavefront.

        Returns
        -------
        position : float
            The position of the `Wavefront` from its starting point
            in metres.
        """
        return self.position

    def get_phase_radius(self: Wavefront) -> float:
        """
        Accessor for the phase radius of the wavefront.

        Returns
        -------
        phase_radius : float
            The phase radius of the wavefront. This is a unitless
            quantity.
        """
        return self.phase_radius

    def get_pixel_positions(self: Wavefront) -> Array:
        """
        Returns
        -------
        positions : Tensor, metres
            The position of each pixel aligned according to the `fft`
            algorithm that is implemented by `numpy`.
        """
        pixels = self.number_of_pixels()
        positions = np.array(np.indices((pixels, pixels)) - pixels / 2.0)
        return self.get_pixel_scale() * positions

    def is_angular(self: Wavefront) -> bool:
        """ """
        return self.angular

    # NOTE: The pixel scale cannot be set when self.angular == True
    # NOTE: This has the correct units always/
    def get_pixel_scale(self: Wavefront):
        """

        NOTE - This seems dodgy, because if we enforce only using
        getter and setter this would result in infinite recursion.
        Maybe this is okay though because it IS The getter

        Returns
        -------
        pixel_scale : The width of a single pixel in the array
            representing the `Wavefront`. Note that this differs
            from `poppy` because it includes the oversample.
        """
        return jax.lax.cond(
            self.is_angular(),
            lambda: self.pixel_scale / self.focal_length,
            lambda: self.pixel_scale,
        )

    def get_waist_radius(self: Wavefront):
        """
        Calculates the current waist radius as half of the diameter

        Mainly just exists to keep consistency with textbook terms
        """
        return self.waist_radius

    def rayleigh_distance(self: Wavefront) -> Scalar:
        """
        Calculates the rayleigh distance of the Gaussian beam.

        Returns
        -------
        rayleigh_distance : float
            The Rayleigh distance of the wavefront in metres.
        """
        return np.pi * self.get_waist_radius() ** 2 / self.get_wavelength()

    # Setters #
    # NOT USED
    def set_position(self: Wavefront, position: float) -> Wavefront:
        """
        Mutator for the position of the wavefront. Changes the
        pixel_scale which is a function of the position.

        Parameters
        ----------
        position : float
            The new position of the wavefront from its starting point
            assumed to be in metres.

        Returns
        -------
        wavefront : Wavefront
            This wavefront at the new position.
        """
        return eqx.tree_at(
            lambda wavefront: wavefront.position,
            self,
            position,
            is_leaf=lambda leaf: leaf is None,
        )

    # NOT USED
    def set_phase_radius(self: Wavefront, phase_radius: float) -> Wavefront:
        """
        Mutator for the phase_radius.

        Parameters
        ----------
        phase_radius : float
            The new phase_radius in metres.

        Returns
        -------
        wavefront : Wavefront
            A modified Wavefront with the new phase_radius.
        """
        return eqx.tree_at(
            lambda wavefront: wavefront.phase_radius,
            self,
            phase_radius,
            is_leaf=lambda leaf: leaf is None,
        )

    # NOTE: ordering convention: dunder, _..., ..._at, ..._after,
    # set_... get_...
    # NOTE: naming convention: self._outside_to_outside position -> absolute
    # place on optical
    # axis and distance -> movement.
    def set_waist_position(
        self: Wavefront, waist_position: Scalar
    ) -> Wavefront:
        """
        Parameters
        ----------
        waist_position : float, metres
            The absolute position of the waist of the wavefront.

        Returns
        -------
        wavefront : Wavefront
            The `Wavefront` with the updated parameters.
        """
        return eqx.tree_at(
            lambda wave: wave.waist_position, self, waist_position
        )

    def set_waist_radius(self: Wavefront, waist_radius: Scalar) -> Wavefront:
        """
        Parameters
        ----------
        beam_radius : float
            The new beam_radius in metres.

        Returns
        -------
        wavefront : Wavefront
            A modified `Wavefront` with the new `beam_radius`.
        """
        return eqx.tree_at(lambda wave: wave.waist_radius, self, waist_radius)

    def set_spherical(self: Wavefront, spherical: bool) -> Wavefront:
        """
        Parameters
        ----------
        spherical : bool
            Toggle the state of the `Wavefront` to and from `spherical`.

        Returns
        -------
        wavefront : Wavefront
            The `Wavefront` with the parameters modified.
        """

        # Pretty sure the 'spherical' input need to wrapped into a jnp array
        return eqx.tree_at(lambda wave: wave.spherical, self, spherical)

    # TODO: Do I even want to include this functionality as it is
    # only for the translation between the angular and the physical
    # coordinates. I'm not sure that we need this.
    def set_focal_length(self: Wavefront, focal_length: Scalar) -> Wavefront:
        """
        Parameters
        ----------
        focal_length : float, metres
            The `focal_length` of the `Wavefront`.

        Returns
        -------
        wavefront : Wavefront
            The `Wavefront` with the parameters modified.
        """
        return eqx.tree_at(lambda wave: wave.focal_length, self, focal_length)

    # NOT USED
    def calculate_pixel_scale(self: Wavefront, position: float) -> None:
        """
        The pixel scale at the position along the axis of propagation.

        Parameters
        ----------
        position : float
            The position of the wavefront along the axis of propagation
            in metres.
        """
        new_pixel_scale = (
            self.get_wavelength()
            * np.abs(position)
            / self.number_of_pixels()
            / self.get_pixel_scale()
        )
        return new_pixel_scale

    # NOT USED
    def is_inside(self: Wavefront, distance: float) -> bool:
        """
        Determines whether a point at along the axis of propagation
        at distance away from the current position is inside the
        rayleigh distance.

        Parameters
        ----------
        distance : float
            The distance to test in metres.

        Returns
        -------
        inside : bool
            true if the point is within the rayleigh distance false
            otherwise.
        """
        return (
            np.abs(self.get_position() + distance - self.waist_position)
            <= self.rayleigh_distance()
        )

    # NOTE: This also needs an ..._after name. I could use something
    # like quadratic_phase_after() or phase_after()
    def quadratic_phase(self: Wavefront, distance: Scalar) -> Array:
        """
        Convenience function that simplifies many of the diffraction
        equations. Calculates a quadratic phase factor associated with
        Parameters
        ----------
        distance : float
            The distance of the propagation measured in metres.

        Returns
        -------
        phase : float
            The near-field quadratic phase accumulated by the beam
            from a propagation of distance.
        """
        positions = self.get_pixel_positions()
        rho_squared = (positions**2).sum(axis=0)
        return np.exp(1.0j * np.pi * rho_squared / distance / self.wavelength)

    def transfer(self: Wavefront, distance: Scalar) -> Array:
        """
        The optical transfer function (OTF) for the Gaussian beam.
        Assumes propagation is along the axis.

        Parameters
        ----------
        distance : float
            The distance to propagate the wavefront along the beam
            via the optical transfer function in metres.

        Returns
        -------
        phase : float
            A phase representing the evolution of the wavefront over
            the distance.
        """
        positions = self.get_pixel_positions()
        x, y = positions[0], positions[1]
        rho_squared = (
            x / (self.get_pixel_scale() ** 2 * self.number_of_pixels())
        ) ** 2 + (
            y / (self.get_pixel_scale() ** 2 * self.number_of_pixels())
        ) ** 2
        # Transfer Function of diffraction propagation eq. 22, eq. 87
        return np.exp(-1.0j * np.pi * self.wavelength * distance * rho_squared)

    def curvature_at(self: Wavefront, position: float) -> float:
        """
        Calculate the radius of curvature of the `Wavefront` phase
        at the absolute position: `position`.

        Parameters
        ----------
        position : float, metres
            The absolute position of the wave along the optical axis
            from spawn.

        Returns
        -------
        radius_of_curvature : float, radians
            The radius of phase curvature for the wavefront.
        """
        relative_position = position - self.waist_position
        return (
            relative_position
            + self.rayleigh_distance() ** 2 / relative_position
        )

    def radius_at(self: Wavefront, position: Scalar) -> Scalar:
        """
        Calculate the radius of the `Wavefront` at an absolute
        position.

        Parameters
        ----------
        position : float, metres
            The absolute position of the `Wavefront` since spawn.

        Returns
        -------
        radius : float, metres
            The radius of the beam.
        """
        relative_position = position - self.waist_position
        return self.get_waist_radius() * np.sqrt(
            1.0 + (relative_position / self.rayleigh_distance()) ** 2
        )

    def is_planar_at(self: Wavefront, position: Scalar) -> bool:
        """
        Determines whether a point at along the axis of propagation
        at distance away from the current position is inside the
        rayleigh distance.

        Parameters
        ----------
        distance : float
            The distance to test in metres.

        Returns
        -------
        inside : bool
            true if the point is within the rayleigh distance false
            otherwise.
        """
        return (
            np.abs(self.waist_position - position) < self.rayleigh_distance()
        )

    # NOTE: Also updates, so I want better names for these rather than
    # after.
    # NOTE: This is only for transitions from planar to spherical
    # or vice versa so it needs a much better name than current.
    def pixel_scale_after(self: Wavefront, distance: float) -> Wavefront:
        """
        Calculate and assign the pixel scale of the `Wavefront` after
        travelling distance. Note that this transformation is dependent
        on the mode of propagation and is only correct for
        `_spherical_to_waist` and `_waist_to_spherical` but not for
        `_plane_to_plane`.

        Parameters
        ----------
        distance : float, metres
            The distance of propagation from the current position.

        Returns
        -------
        wavefront : Wavefront
            The wavefront but the pixel_scale has been updated.
        """
        pixel_scale = (
            self.get_wavelength()
            * np.abs(distance)
            / (self.number_of_pixels() * self.get_pixel_scale())
        )
        return self.set_pixel_scale(pixel_scale)

    def position_after(self: Wavefront, distance: Scalar) -> Wavefront:
        """
        Move the wavefront forward by `distance`.

        Parameters
        ----------
        distance : float, metres
            The distance of propagation.

        Returns
        -------
        wavefront : Wavefront
            The `Wavefront` with the `position` leaf updated.
        """
        position = self.position + distance
        return eqx.tree_at(lambda wave: wave.position, self, position)


class GaussianPropagator(eqx.Module):
    distance: float

    def __init__(self: Propagator, distance: float):
        self.distance = np.asarray(distance).astype(float)

    def _fourier_transform(self: Propagator, field: Array) -> Array:
        return 1 / field.shape[0] * np.fft.fft2(field)

    def _inverse_fourier_transform(self: Propagator, field: Array) -> Array:
        return field.shape[0] * np.fft.ifft2(field)

    # NOTE: need to add in the standard FFT normalising factor
    # as in the propagator.
    def _propagate(self: Propagator, field: Array, distance: float) -> Array:
        return jax.lax.cond(
            distance > 0,
            lambda: self._inverse_fourier_transform(field),
            lambda: self._fourier_transform(field),
        )

    # NOTE: Wavefront must be planar
    # NOTE: Uses eq. 82, 86, 87
    def _plane_to_plane(
        self: Propagator, wavefront: Wavefront, distance: float
    ):
        # NOTE: Seriously need to change the name to get_field()
        field = self._fourier_transform(wavefront.get_complex_form())
        field *= np.fft.fftshift(wavefront.transfer(distance))  # eq. 6.68
        field = self._inverse_fourier_transform(field)

        wavefront = (
            wavefront.position_after(distance)
            .set_phase(np.angle(field))
            .set_amplitude(np.abs(field))
        )
        return wavefront

    # NOTE: I'm thinking that the logic for repacking the wavefront
    # should occur somewhere else although I guess that it can't really
    # NOTE: Must start with a planar wavefront
    def _waist_to_spherical(
        self: Propagator, wavefront: Wavefront, distance: float
    ) -> Wavefront:
        # Lawrence eq. 83,88
        field = wavefront.get_complex_form()
        field *= np.fft.fftshift(
            wavefront.quadratic_phase(distance)
        )  # Wavelength dependent
        field = self._propagate(
            field, distance
        )  # Wrapper for forwards/reverse FFTs
        wavefront = (
            wavefront.pixel_scale_after(distance)
            .position_after(distance)
            .set_phase(np.angle(field))
            .set_amplitude(np.abs(field))
            .set_spherical(True)
        )
        return wavefront

    # Wavefront.spherical must be True initially
    def _spherical_to_waist(
        self: Propagator, wavefront: Wavefront, distance: float
    ) -> Wavefront:
        # Lawrence eq. 89
        field = wavefront.get_complex_form()
        field = self._propagate(field, distance)
        wavefront = wavefront.pixel_scale_after(distance)
        field *= np.fft.fftshift(
            wavefront.quadratic_phase(distance)
        )  # Wavelength dependent

        wavefront = (
            wavefront.set_phase(np.angle(field))
            .set_amplitude(np.abs(field))
            .set_spherical(True)
            .position_after(distance)
        )
        return wavefront

    def _inside_to_inside(self: Propagator, wave: Wavefront) -> Wavefront:
        wave = self._plane_to_plane(wave, self.distance)
        return wave

    def _inside_to_outside(self: Propagator, wave: Wavefront) -> Wavefront:
        start = wave.position
        end = wave.position + self.distance
        wave = self._plane_to_plane(wave, wave.waist_position - start)
        wave = self._waist_to_spherical(wave, end - wave.waist_position)
        return wave

    def _outside_to_inside(self: Propagator, wave: Wavefront) -> Wavefront:
        start = wave.position
        end = wave.position + self.distance
        wave = self._spherical_to_waist(wave, wave.waist_position - start)
        wave = self._plane_to_plane(wave, end - wave.waist_position)
        return wave

    def _outside_to_outside(self: Propagator, wave: Wavefront) -> Wavefront:
        start = wave.position
        end = wave.position + self.distance
        wave = self._spherical_to_waist(wave, wave.waist_position - start)
        wave = self._waist_to_spherical(wave, end - wave.waist_position)
        return wave

    # NOTE: So I could attempt to move all of the functionality into
    # the wavefront class and do very little here. Damn, I need to
    # fit it into the overall architecture.
    # TODO: Implement the oversample in the fixed sampling propagator
    # Coordinates must be in metres for the propagator
    def __call__(self: Propagator, wave: Wavefront) -> Wavefront:
        # NOTE: need to understand this mystery.
        field = np.fft.fftshift(wave.get_complex_form())
        wave = wave.update_phasor(np.abs(field), np.angle(field))
        position = wave.position + self.distance
        decision = 2 * wave.spherical + wave.is_planar_at(position)

        wave = jax.lax.switch(
            decision,
            [
                self._inside_to_outside,
                self._inside_to_inside,
                self._outside_to_outside,
                self._outside_to_inside,
            ],
            wave,
        )

        field = np.fft.fftshift(wave.get_complex_form())
        wave = wave.update_phasor(np.abs(field), np.angle(field))
        return wave


class GaussianLens(eqx.Module):
    focal_length: float
    # TODO: Should this store its position in the optical system?
    # No I don't think that it should.

    def __init__(self: Layer, focal_length: float) -> Layer:
        self.focal_length = np.asarray(focal_length).astype(float)

    def _phase(self: Layer, wave: Wavefront, distance: float) -> Array:
        position = wave.get_pixel_positions()
        rho_squared = (position**2).sum(axis=0)
        return np.exp(1.0j / np.pi * rho_squared / distance * wave.wavelength)

    def __call__(self: Layer, wave: Wavefront) -> Wavefront:
        from_waist = wave.waist_position - wave.position
        was_spherical = (
            np.abs(from_waist)
            > wave.rayleigh_factor * wave.rayleigh_distance()
        )

        curve = wave.curvature_at(wave.position)

        curve_at = jax.lax.cond(
            was_spherical, lambda: from_waist, lambda: np.inf
        )

        curve_after = jax.lax.cond(
            was_spherical,
            lambda: 1.0 / (1.0 / curve - 1.0 / self.focal_length),
            lambda: -self.focal_length,
        )

        radius = wave.radius_at(wave.position)
        curve_ratio = (
            wave.wavelength * curve_after / np.pi / radius**2
        ) ** 2
        curve_matched = curve == self.focal_length

        waist_position_after = jax.lax.cond(
            curve_matched,
            lambda: wave.position,
            lambda: -curve_after / (1.0 + curve_ratio) + wave.position,
        )

        waist_radius_after = jax.lax.cond(
            curve_matched,
            lambda: radius,
            lambda: radius / np.sqrt(1.0 + 1.0 / curve_ratio),
        )

        focal_length_after = jax.lax.cond(
            np.isinf(wave.focal_length),
            lambda: self.focal_length,
            lambda: curve_after / curve_at * wave.focal_length,
        )

        wave = (
            wave.set_focal_length(focal_length_after)
            .set_waist_radius(waist_radius_after)
            .set_waist_position(waist_position_after)
        )

        from_new_waist = waist_position_after - wave.position
        is_spherical = np.abs(from_new_waist) > wave.rayleigh_distance()

        distance = 1.0 / jax.lax.cond(
            wave.spherical,
            lambda: jax.lax.cond(
                is_spherical,
                lambda: 1.0 / self.focal_length
                + 1.0 / from_new_waist
                - 1.0 / curve_at,
                lambda: 1.0 / self.focal_length - 1.0 / curve_at,
            ),
            lambda: jax.lax.cond(
                is_spherical,
                lambda: 1.0 / self.focal_length + 1.0 / from_new_waist,
                lambda: 1.0 / self.focal_length,
            ),
        )

        field = wave.get_complex_form() * self._phase(wave, distance)
        phase = np.angle(field)
        amplitude = np.abs(field)

        return (
            wave.set_phase(phase)
            .set_amplitude(amplitude)
            .set_spherical(is_spherical)
            .set_waist_position(waist_position_after)
            .set_waist_radius(waist_radius_after)
            .set_focal_length(focal_length_after)
        )



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import (
    Telescope,
    Dither,
    LayeredOpticalSystem,
    LayeredDetector,
    PointSource,
    PSF,
    Optic,
    AddConstant,
)


@pytest.fixture
def optics():
    return LayeredOpticalSystem(16, 1.0, [Optic()])


@pytest.fixture
def detector():
    return LayeredDetector([AddConstant(1)])


@pytest.fixture
def source():
    return PointSource([1e-6])


def _test_model(inst):
    assert isinstance(inst.model(), np.ndarray)
    assert isinstance(inst.model(return_psf=True), PSF)


def test_instrument(optics, detector, source):
    # All inputs
    _test_model(Telescope(optics, source, detector))

    # No detector
    _test_model(Telescope(optics, source))

    # Single source with key
    _test_model(Telescope(optics, ("source", source)))

    # Multiple sources
    _test_model(Telescope(optics, [source, source]))

    # Test failures
    with pytest.raises(TypeError):
        Telescope(1, source, detector)
    with pytest.raises(TypeError):
        Telescope(optics, source, 1)

    # Test getattr
    inst = Telescope(optics, source, detector)
    inst.opd
    with pytest.raises(AttributeError):
        inst.not_an_attr


def test_dither(optics, detector, source):
    _test_model(Dither(np.zeros((1, 2)), optics, source, detector))

    with pytest.raises(ValueError):
        Dither(np.zeros((1, 3)), optics, source, detector)



from __future__ import annotations
from typing import Union

__all__ = ["Rotate", "Flip", "Resize"]


from .optical_layers import OpticalLayer
from .detector_layers import DetectorLayer
from ..wavefronts import Wavefront
from ..psfs import PSF


class UnifiedLayer(OpticalLayer, DetectorLayer):
    """
    Base class for unified layers that can be applied to either wavefronts or PSFs.
    """


class Resize(UnifiedLayer):
    """
    Resizes either a wavefront or PSF by either padding or cropping. Note this class
    only supports padding and cropping of even sizes to even sizes, and odd sizes to
    odd sizes to ensure all operations are paraxial.

    ??? abstract "UML"
        ![UML](../../assets/uml/Resize.png)

    Attributes
    ----------
    npixels : int
        The desired output size.
    """

    npixels: int

    def __init__(self: UnifiedLayer, npixels: int):
        """
        Parameters
        ----------
        npixels : tuple
            The desired output size.
        """
        super().__init__()
        self.npixels = int(npixels)

    def apply(
        self: UnifiedLayer, input: Union[Wavefront, PSF]
    ) -> Union[Wavefront, PSF]:
        """
        Resizes the input.

        Parameters
        ----------
        input : Union[Wavefront, PSF]
            The input to resize.

        Returns
        -------
        input : Union[Wavefront, PSF]
            The resized input.
        """
        return input.resize(self.npixels)


class Rotate(UnifiedLayer):
    """
    Rotates either a wavefront or PSF by a given angle. This is done using
    interpolation methods. The 'complex' input only has an effect if the input is a
    wavefront.

    ??? abstract "UML"
        ![UML](../../assets/uml/Rotate.png)

    Attributes
    ----------
    angle : float, radians
        The angle by which to rotate the input in the clockwise direction.
    order : int
        The order of the interpolation to use. Must be 0 or 1.
    complex : bool
        Should the rotation be performed on the 'complex' (real, imaginary), as opposed
        to the default 'phasor' (amplitude, phase) arrays. Only applies if the input is
        a wavefront.
    """

    angle: float
    order: int
    complex: bool

    def __init__(
        self: UnifiedLayer,
        angle: float,
        order: int = 1,
        complex: bool = False,
    ):
        """
        Parameters
        ----------
        angle : float, radians
            The angle by which to rotate the input in the clockwise direction.
        order : int = 1
            The order of the interpolation to use. Must be 0 or 1.
        complex : bool = False
            Should the rotation be performed on the 'complex' (real, imaginary), as
            opposed to the default 'phasor' (amplitude, phase) arrays. Only applies if
            the input is a wavefront.
        """
        super().__init__()
        self.angle = float(angle)
        self.order = int(order)
        self.complex = bool(complex)

        if self.order not in (0, 1):
            raise ValueError("Order must be 0, 1")

    def apply(
        self: UnifiedLayer, input: Union[Wavefront, PSF]
    ) -> Union[Wavefront, PSF]:
        """
        Applies the rotation to the input.

        Parameters
        ----------
        input : Union[Wavefront, PSF]
            The input to rotate.

        Returns
        -------
        input : Union[Wavefront, PSF]
            The rotated input.
        """
        if isinstance(input, PSF):
            return input.rotate(self.angle, self.order)
        else:
            return input.rotate(self.angle, self.order, self.complex)


class Flip(UnifiedLayer):
    """
    Flips either a wavefront or PSF about the input axes. Can be either an int or a
    tuple of ints. This class uses the 'ij' indexing convention, ie axis 0 is the
    y-axis, and axis 1 is the x-axis.

    ??? abstract "UML"
        ![UML](../../assets/uml/Flip.png)

    Attributes
    ----------
    axes : Union[tuple, int]
        The axes to flip the input about. This class uses the 'ij' indexing convention,
        ie axis 0 is the y-axis, and axis 1 is the x-axis.
    """

    axes: Union[tuple[int], int]

    def __init__(self: UnifiedLayer, axes: Union[tuple[int], int]):
        """
        Parameters
        ----------
        axes : Union[tuple, int]
            The axes to flip the input about. This class uses the 'ij' indexing
            convention, ie axis 0 is the y-axis, and axis 1 is the x-axis.
        """
        super().__init__()
        self.axes = axes

        if isinstance(self.axes, tuple):
            for axes in self.axes:
                if not isinstance(axes, int):
                    raise ValueError("All axes must be integers.")
        elif not isinstance(self.axes, int):
            raise ValueError("axes must be integers.")

    def apply(
        self: UnifiedLayer, input: Union[Wavefront, PSF]
    ) -> Union[Wavefront, PSF]:
        """
        Flips the input about the input axes.

        Parameters
        ----------
        input : Union[Wavefront, PSF]
            The input to flip.

        Returns
        -------
        input : Union[Wavefront, PSF]
            The flipped input.
        """
        return input.flip(self.axes)



from jax import numpy as np, config

config.update("jax_debug_nans", True)
import pytest
from dLux import Spectrum, PolySpectrum


def test_spectrum():
    # Test 1d
    Spectrum(np.ones(3))
    spec = Spectrum(np.ones(3), np.ones(3))
    assert isinstance(spec.normalise(), Spectrum)

    # Test 2d
    spec = Spectrum(np.ones(3), np.ones((2, 3)))
    assert isinstance(spec.normalise(), Spectrum)

    # Test failure
    with pytest.raises(ValueError):
        Spectrum(np.ones((2, 3)), np.ones(3))
    with pytest.raises(ValueError):
        Spectrum(np.ones(3), np.ones((2, 4)))


def test_poly_spectrum():
    spec = PolySpectrum(np.ones(3), np.ones(3))
    assert isinstance(spec.weights, np.ndarray)
    assert isinstance(spec.normalise(), PolySpectrum)

    with pytest.raises(ValueError):
        PolySpectrum(np.ones(3), np.ones((2, 1)))



from __future__ import annotations
from zodiax import Base
import jax.numpy as np
from jax import Array
import jax.tree as jtu
import dLux.utils as dlu

__all__ = ["Zernike", "ZernikeBasis"]


# TODO: Should all the leaves of this class be static?
class Zernike(Base):
    """
    A Zernike polynomial that can be generated dynamically in a way that is both jit and
    grad safe. If you want a _static_ zernike (most use cases), use the zernike
    functions found in `utils.zernikes` and load the basis into a `BasisOptic` class.

    The 'jth' zernike polynomial is defined [here](https://oeis.org/A176988). The basic
    translation between the noll index and the pair of numbers is shown below:

    1 -> (0, 0) : Piston

    2, 3 -> (1, -1), (1, 1) : Tip, Tilt

    4, 5, 6 -> (2, -2), (2, 0), (2, 2) : Defocus, Astigmatism

    7, 8, 9, 10 -> (3, -3), (3, -1), (3, 1), (3, 3) : Coma, Trefoil

    ??? abstract "UML"
        ![UML](../../assets/uml/Zernike.png)

    Attributes
    ----------
    j : int
        The Zernike (noll) index.
    n : int
        The radial order of the Zernike polynomial.
    m : int
        The azimuthal order of the Zernike polynomial.
    name : str
        The name of the Zernike polynomial.
    _c : Array
        The array of normalisation coefficients used in the radial calculation.
        This is a pre-calculated parameter and should not be changed.
    _k : Array
        The array of powers using the radial calculation. This is a pre-calculated
        parameter and should not be changed.
    """

    j: int
    n: int
    m: int
    name: str
    _c: Array
    _k: Array

    def __init__(self: Zernike, j: int):
        """
        Construct for the Zernike class.

        Parameters
        ----------
        j : int
            The Zernike (noll) index.
        """
        if int(j) < 1:
            raise ValueError("The Zernike index must be greater than 0.")
        self.j = int(j)
        self.name = dlu.zernike_name(j)
        self.n, self.m = dlu.noll_indices(j)
        self._c, self._k = dlu.zernike_factors(j)

    def calculate(self: Zernike, coordinates: Array, nsides: int = 0) -> Array:
        """
        Calculates the Zernike polynomial.

        Note: standard zernike polynomials are only defined up to a radial value of 1,
        so generating one that spans the entire aperture needs a diameter of 2.

        Parameters
        ----------
        coordinates : Array
            The Cartesian coordinates to calculate the Zernike upon.
        nsides : int
            The number of sides of the aperture. If 0, the Zernike is calculated
            on a circular aperture.

        Returns
        -------
        zernike : Array
            The Zernike polynomial.
        """
        if nsides == 0:
            return dlu.zernike_fast(
                self.n, self.m, self._c, self._k, coordinates
            )
        else:
            return dlu.polike_fast(
                nsides, self.n, self.m, self._c, self._k, coordinates
            )


# TODO: Rename basis to basis_fns??
class ZernikeBasis(Base):
    """
    A Basis of Zernike polynomials that can be generated dynamically in a way that is
    both jit and grad safe. If you want a _static_ zernike (most use cases), use the
    zernike functions found in `utils.zernikes` and load the basis into a `BasisOptic`
    class.

    The 'jth' zernike polynomial is defined [here](https://oeis.org/A176988). The basic
    translation between the noll index and the pair of numbers is shown below:

    1 -> (0, 0) : Piston

    2, 3 -> (1, -1), (1, 1) : Tip, Tilt

    4, 5, 6 -> (2, -2), (2, 0), (2, 2) : Defocus, Astigmatism

    7, 8, 9, 10 -> (3, -3), (3, -1), (3, 1), (3, 3) : Coma, Trefoil

    ??? abstract "UML"
        ![UML](../../assets/uml/ZernikeBasis.png)

    Attributes
    ----------
    basis : list[Zernike]
        The list of `Zernike` polynomial classes to calculate.
    """

    basis: list[Zernike]

    def __init__(self: ZernikeBasis, js: list[int]):
        """
        Constructor for the DynamicZernike class.

        Parameters
        ----------
        js : list[int]
            The list of Zernike (noll) indices to calculate.
        """
        self.basis = [Zernike(j) for j in js]

    def calculate_basis(
        self: ZernikeBasis, coordinates: Array, nsides: int = 0
    ) -> Array:
        """
        Calculates the full Zernike polynomial basis.

        Note: standard zernike polynomials are only defined up to a radial value of 1,
        so generating a basis that spans the entire aperture needs a diameter of 2.

        Parameters
        ----------
        coordinates : Array
            The Cartesian coordinates to calculate the Zernike basis upon.
        nsides : int
            The number of sides of the aperture. If 0, the Zernike basis is calculated
            on a circular aperture.

        Returns
        -------
        basis : Array
            The Zernike polynomial basis.
        """
        leaf_fn = lambda leaf: isinstance(leaf, Zernike)
        calculate_fn = lambda z: z.calculate(coordinates, nsides)
        return np.array(jtu.map(calculate_fn, self.basis, is_leaf=leaf_fn))



import jax.numpy as np
from jax import Array, vmap
import dLux.utils as dlu


__all__ = ["FFT", "MFT", "fresnel_MFT"]


def FFT(
    phasor: Array,
    wavelength: float,
    pixel_scale: float,
    focal_length: float = None,
    pad: int = 2,
    inverse: bool = False,
) -> Array:
    """
    Calculates the Fast Fourier Transform (FFT) of the input phasor.

    Parameters
    ----------
    phasor : Array[complex]
        The input phasor.
    wavelength : float, meters
        The wavelength of the input phasor.
    pixel_scale : float, meters/pixel
        The pixel scale of the input phasor.
    focal_length : float = None
        The focal length of the propagation. If None, the output pixel scale has units
        of radians, else meters.
    pad : int = 2
        The amount to pad the input array by before propagation. Note this function
        does not automatically crop the output.
    inverse : bool = False
        Is this a forward or inverse FFT.

    Returns
    -------
    phasor : Array[complex]
        The propagated phasor.
    """
    npixels = phasor.shape[-1]

    # Calculate the output pixel scale
    fringe_size = wavelength / (pixel_scale * npixels)
    new_pixel_scale = fringe_size / pad
    if focal_length is not None:
        new_pixel_scale *= focal_length

    # Pad the input array
    npixels = (npixels * (pad - 1)) // 2
    phasor = np.pad(phasor, npixels)

    # Perform the FFT
    if inverse:
        phasor = np.fft.fft2(np.fft.ifftshift(phasor)) / phasor.shape[-1]
    else:
        phasor = np.fft.fftshift(np.fft.ifft2(phasor)) * phasor.shape[-1]

    return phasor, new_pixel_scale


def transfer_matrix(
    wavelength: float,
    npixels_in: int,
    pixel_scale_in: float,
    npixels_out: int,
    pixel_scale_out: float,
    shift: float = 0.0,
    focal_length: float = None,
    focal_shift: float = 0.0,
    inverse: bool = False,
) -> Array:
    """
    Calculates the transfer matrix for the MFT.

    Parameters
    ----------
    wavelength : float, meters
        The wavelength of the input phasor.
    npixels_in : int
        The number of pixels in the input plane.
    pixel_scale_in : float, meters/pixel, radians/pixel
        The pixel scale of the input plane.
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : float, meters/pixel or radians/pixel
        The pixel scale of the output plane.
    shift : float = 0.0
        The shift in the center of the output plane.
    focal_length : float = None
        The focal length of the propagation. If None, the propagation is angular and
        pixel_scale_out is taken in as radians/pixel, else meters/pixel.
    focal_shift: float, meters
        The shift from focus to propagate to. Used for fresnel propagation.
    inverse: bool = False
        Is this a forward or inverse propagation.

    Returns
    -------
    transfer_matrix : Array
        The transfer matrix for the MFT.
    """
    # Get parameters
    fringe_size = wavelength / (pixel_scale_in * npixels_in)

    # Input coordinates
    scale_in = 1.0 / npixels_in
    in_vec = dlu.nd_coords(npixels_in, scale_in, shift * scale_in)

    # Output coordinates
    scale_out = pixel_scale_out / fringe_size
    if focal_length is not None:
        # scale_out /= focal_length
        scale_out /= focal_length + focal_shift
    out_vec = dlu.nd_coords(npixels_out, scale_out, shift * scale_out)

    # Generate transfer matrix
    matrix = 2j * np.pi * np.outer(in_vec, out_vec)
    if inverse:
        matrix *= -1
    return np.exp(matrix)


def calc_nfringes(
    wavelength: float,
    npixels_in: int,
    pixel_scale_in: int,
    npixels_out: int,
    pixel_scale_out: float,
    focal_length: float = None,
    focal_shift: float = 0.0,
) -> Array:
    """
    Calculates the number of fringes in the output plane.

    Parameters
    ----------
    wavelength : float, meters
        The wavelength of the input phasor.
    npixels_in : int
        The number of pixels in the input plane.
    pixel_scale_in : float, meters/pixel, radians/pixel
        The pixel scale of the input plane.
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : float, meters/pixel or radians/pixel
        The pixel scale of the output plane.
    focal_length : float = None
        The focal length of the propagation. If None, the propagation is angular and
        pixel_scale_out is taken in as radians/pixel, else meters/pixel.
    focal_shift: float, meters
        The shift from focus to propagate to. Used for fresnel propagation.

    Returns
    -------
    nfringes : Array
        The number of fringes in the output plane.
    """
    # Fringe size
    diameter = npixels_in * pixel_scale_in
    fringe_size = wavelength / diameter

    # Output array size
    output_size = npixels_out * pixel_scale_out
    if focal_length is not None:
        output_size /= focal_length + focal_shift

    # Fringe size and number of fringes
    return output_size / fringe_size


def MFT(
    phasor: Array,
    wavelength: float,
    pixel_scale_in: float,
    npixels_out: int,
    pixel_scale_out: float,
    focal_length: float = None,
    shift: Array = np.zeros(2),
    pixel: bool = True,
    inverse: bool = False,
) -> Array:
    """
    Propagates a phasor using a Matrix Fourier Transform (MFT), allowing for output
    pixel scale and a shift to be specified.

    This Soumer et al. 2007 paper describes the MFT: https://arxiv.org/pdf/0711.0368

    Parameters
    ----------
    phasor : Array
        The input phasor.
    wavelength : float, meters
        The wavelength of the input phasor.
    pixel_scale_in : float, meters/pixel, radians/pixel
        The pixel scale of the input plane.
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : float, meters/pixel or radians/pixel
        The pixel scale of the output plane.
    focal_length : float = None
        The focal length of the propagation. If None, the propagation is angular and
        pixel_scale_out is taken in as radians/pixel, else meters/pixel.
    shift : Array = np.zeros(2)
        The shift in the center of the output plane.
    pixel : bool = True
        Should the shift be taken in units of pixels, or pixel scale.


    Returns
    -------
    phasor : Array
        The propagated phasor.
    """
    # Get parameters
    npixels_in = phasor.shape[-1]
    if not pixel:
        shift /= pixel_scale_out

    # Alias the transfer matrix function
    get_tf_mat = lambda s: transfer_matrix(
        wavelength,
        npixels_in,
        pixel_scale_in,
        npixels_out,
        pixel_scale_out,
        s,
        focal_length,
        0.0,
        inverse,
    )

    # Get transfer matrices and propagate
    x_mat, y_mat = vmap(get_tf_mat)(shift)
    phasor = (y_mat.T @ phasor) @ x_mat

    # Normalise
    nfringes = calc_nfringes(
        wavelength,
        npixels_in,
        pixel_scale_in,
        npixels_out,
        pixel_scale_out,
        focal_length,
    )
    phasor *= np.exp(
        np.log(nfringes) - (np.log(npixels_in) + np.log(npixels_out))
    )

    return phasor


# Move to utils as thinlens?
def quadratic_phase(
    wavelength: float,
    distance: float,
    coordinates: Array,
) -> Array:
    """
    A function to calculate quadratic phase factors, used for fresnel propagation.

    Parameters
    ----------
    wavelength : float, meters
        The wavelength of the input phasor.
    distance : float, meters
        The 'focal distance' of the quadratic lens.
    coordinates : Array
        The coordinates to calculate the phase factors for.

    Returns
    -------
    phase : Array
        The phases for a quadratic lens.
    """
    return np.exp(
        1j * np.pi * np.hypot(*coordinates) ** 2 / (distance * wavelength)
    )


def fresnel_phase_factors(
    wavelength: float,
    npixels_in: int,
    pixel_scale_in: float,
    npixels_out: int,
    pixel_scale_out: float,
    focal_length: float,
    focal_shift: float,
) -> tuple:
    """
    Calculates the phase factors for the Fresnel propagation.

    Parameters
    ----------
    wavelength : float, meters
        The wavelength of the input phasor.
    npixels_in : int
        The number of pixels in the input plane.
    pixel_scale_in : float, meters/pixel, radians/pixel
        The pixel scale of the input plane.
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : float, meters/pixel or radians/pixel
        The pixel scale of the output plane.
    focal_length : float
        The focal length of the propagation.
    focal_shift: float, meters
        The shift from focus to propagate to.

    Returns
    -------
    first_factor : Array
        The first factor in the fresnel propagation.
    second_factor : Array
        The second factor in the fresnel propagation.
    """
    # Calculate parameters
    prop_dist = focal_length + focal_shift
    input_positions = dlu.pixel_coords(npixels_in, pixel_scale_in)
    output_positions = dlu.pixel_coords(npixels_out, pixel_scale_out)

    # Calculate phase factors
    phase_fn = lambda dist, coords: quadratic_phase(wavelength, dist, coords)
    first_factor = phase_fn(-focal_length, input_positions) * phase_fn(
        prop_dist, input_positions
    )
    second_factor = np.exp(2j * np.pi * prop_dist / wavelength) * phase_fn(
        prop_dist, output_positions
    )

    return first_factor, second_factor


def fresnel_MFT(
    phasor: Array,
    wavelength: float,
    pixel_scale_in: float,
    npixels_out: int,
    pixel_scale_out: float,
    focal_length: float,
    focal_shift: float,
    shift: Array = np.zeros(2),
    pixel: bool = True,
    inverse: bool = False,
) -> Array:
    """
    Propagates the phasor using a Far-Field Fresnel propagation. This allows for psfs
    to be better modelled a few wavelengths from the focal plane.

    NOTE: This does have an 'inverse' parameter, however behaviour is not
    guaranteed to be correct when `inverse=True`.

    Parameters
    ----------
    phasor : Array[complex]
        The input phasor.
    wavelength : float, meters
        The wavelength of the input phasor.
    pixel_scale_in : float, meters/pixel, radians/pixel
        The pixel scale of the input plane.
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : float, meters/pixel or radians/pixel
        The pixel scale of the output plane.
    focal_length : float
        The focal length of the propagation.
    focal_shift: float, meters
        The shift from focus to propagate to.
    shift : Array = np.zeros(2)
        The shift in the center of the output plane.
    pixel : bool = True
        Should the shift be taken in units of pixels, or pixel scale.
    inverse: bool = False
        Is this a forward or inverse propagation.

    Returns
    -------
    phasor : Array[complex]
        The propagated phasor.
    """
    # Calculate phase factors
    first_factor, second_factor = fresnel_phase_factors(
        wavelength,
        phasor.shape[-1],
        pixel_scale_in,
        npixels_out,
        pixel_scale_out,
        focal_length,
        focal_shift,
    )

    # Propagate
    phasor *= first_factor
    phasor = MFT(
        phasor,
        wavelength,
        pixel_scale_in,
        npixels_out,
        pixel_scale_out,
        focal_length,
        shift,
        pixel,
        inverse,
    )

    phasor *= second_factor
    return phasor



import jax.numpy as np
from jax import Array
import dLux.utils as dlu


def fourier_rotate(
    array: Array, angle: Array, padding: int = 2
) -> Array:  # pragma: no cover
    """
    Rotates an array by the angle, using a fourier rotation.

    Parameters
    ----------
    array : Array
        The array to rotate.
    angle : Array, radians
        The angle to rotate the array by.
    padding : int = 2
        The amount of fourier padding to use.

    Returns
    -------
    array : Array
        The rotated array.
    """
    raise NotImplementedError("Fourier rotation is under development.")
    # TODO implement
    in_shape = array.shape
    array_shape = np.array(in_shape, dtype=int) + 3
    array = (
        np.full(array_shape, np.nan, dtype=float)
        .at[1 : in_shape[0] + 1, 1 : in_shape[1] + 1]
        .set(array)
    )

    # FFT rotation only work in the -45:+45 range
    # So I need to work out how to determine the quadrant that
    # angle is in and hence the
    # number of required pi/2 rotations and angle in radians.
    half_pi_to_1st_quadrant = angle // (np.pi / 2)
    angle_in_1st_quadrant = -angle + (half_pi_to_1st_quadrant * np.pi / 2)

    array = np.rot90(array, half_pi_to_1st_quadrant).at[:-1, :-1].get()

    width, height = array.shape
    left_corner = int(((padding - 1) / 2.0) * width)
    right_corner = int(((padding + 1) / 2.0) * width)
    top_corner = int(((padding - 1) / 2.0) * height)
    bottom_corner = int(((padding + 1) / 2.0) * height)

    # Make the padded array
    out_shape = (width * padding, height * padding)
    padded_array = (
        np.full(out_shape, np.nan, dtype=float)
        .at[left_corner:right_corner, top_corner:bottom_corner]
        .set(array)
    )

    padded_mask = (
        np.ones(out_shape, dtype=bool)
        .at[left_corner:right_corner, top_corner:bottom_corner]
        .set(np.where(np.isnan(array), True, False))
    )

    # Rotate the mask, to know what part is actually the array
    padded_mask = dlu.rotate(padded_mask, -angle_in_1st_quadrant)

    # Replace part outside the array which are NaN by 0, and go into
    # Fourier space.
    padded_array = np.where(np.isnan(padded_array), 0.0, padded_array)

    uncentered_angular_displacement = np.tan(angle_in_1st_quadrant / 2.0)
    centered_angular_displacement = -np.sin(angle_in_1st_quadrant)

    uncentered_frequencies = np.fft.fftfreq(out_shape[0])
    centered_frequencies = np.arange(-out_shape[0] / 2.0, out_shape[0] / 2.0)

    pi_factor = -2.0j * np.pi * np.ones(out_shape, dtype=float)

    uncentered_phase = np.exp(
        uncentered_angular_displacement
        * ((pi_factor * uncentered_frequencies).T * centered_frequencies).T
    )

    centered_phase = np.exp(
        centered_angular_displacement
        * (pi_factor * centered_frequencies).T
        * uncentered_frequencies
    )

    f1 = np.fft.ifft(
        (np.fft.fft(padded_array, axis=0).T * uncentered_phase).T, axis=0
    )

    f2 = np.fft.ifft(np.fft.fft(f1, axis=1) * centered_phase, axis=1)

    rotated_array = (
        np.fft.ifft((np.fft.fft(f2, axis=0).T * uncentered_phase).T, axis=0)
        .at[padded_mask]
        .set(np.nan)
    )

    return np.real(
        rotated_array.at[
            left_corner + 1 : right_corner - 1,
            top_corner + 1 : bottom_corner - 1,
        ].get()
    ).copy()



from collections import OrderedDict
from typing import Any, Callable
import jax.numpy as np
import jax.tree as jtu
from jax import Array

__all__ = ["map2array", "list2dictionary", "insert_layer", "remove_layer"]


def map2array(fn: Callable, tree: Any, leaf_fn: Callable = None) -> Array:
    """
    Maps a function across a pytree, flattening it and turning it into an
    array.

    Parameters
    ----------
    fn : Callable
        The function to be mapped across the pytree.
    tree : Any
        The pytree to be mapped across.
    leaf_fn : Callable = None
        The function to be used to determine whether a leaf is reached.

    Returns
    -------
    array : Array
        The flattened array of the pytree.
    """
    if leaf_fn is not None:
        return np.array(jtu.flatten(jtu.map(fn, tree, is_leaf=leaf_fn))[0])
    else:
        return np.array(jtu.flatten(jtu.map(fn, tree))[0])


def list2dictionary(
    list_in: list, ordered: bool, allowed_types: tuple = ()
) -> dict:
    """
    Converts some input list to a dictionary. The input list entries can either be
    objects, in which case the keys are taken as the class name, else a (key, object)
    tuple can be used to specify a key.

    If any duplicate keys are found, the key is appended with an index value. ie if two
    of the list entries have the same key 'layer', they will be assigned 'layer_0' and
    layer_1' respectively, depending on their input order in the list.

    Parameters
    ----------
    list_in : list
        The list of objects to be converted into a dictionary.
    ordered : bool
        Whether to return an ordered or regular dictionary.
    allowed_types : tuple
        The allowed types of layers to be included in the dictionary.

    Returns
    -------
    dictionary : dict
        The equivalent dictionary or ordered dictionary.
    """
    # Construct names list and identify repeats
    names, repeats = [], []
    for item in list_in:
        # Check for specified names
        if isinstance(item, tuple):
            # item, name = item
            name, item = item
        else:
            name = item.__class__.__name__

        # Check input types
        if allowed_types != () and not isinstance(item, allowed_types):
            raise TypeError(
                f"Item {name} is not an allowed type, got " f"{type(item)}"
            )

        # Check for Repeats
        if name in names:
            repeats.append(name)
        names.append(name)

    # Get list of unique repeats
    repeats = list(set(repeats))

    # Iterate over repeat names
    for i in range(len(repeats)):
        # Iterate over names list and append index value to name
        idx = 0
        for j in range(len(names)):
            if repeats[i] == names[j]:
                names[j] = names[j] + "_{}".format(idx)
                idx += 1

    # Turn list into Dictionary
    dict_out = OrderedDict() if ordered else {}
    for i in range(len(names)):
        # Check for spaces in names
        if " " in names[i]:
            raise ValueError(f"Names can not contain spaces, got {names[i]}")

        # Add to dict
        if isinstance(list_in[i], tuple):
            # item = list_in[i][0]
            item = list_in[i][1]
        else:
            item = list_in[i]
        dict_out[names[i]] = item
    return dict_out


def insert_layer(layers: dict, layer: Any, index: int, type: Any) -> dict:
    """
    Inserts a layer into a dictionary of layers at a specified index. This function
    calls the list2dictionary function to ensure all keys remain unique. Note that this
    can result in some keys being modified if they are duplicates. The input 'layer'
    can be a tuple of (key, layer) to specify a key, else the key is taken as the
    class name of the layer.

    Parameters
    ----------
    layers : dict
        The dictionary of layers to insert the layer into.
    layer : Any
        The layer to be inserted.
    index : int
        The index at which to insert the layer.
    type : Any
        The type of layer to be inserted. Used for type-checking.

    Returns
    -------
    layers : dict
        The updated dictionary of layers.
    """
    layers_list = list(zip(layers.keys(), layers.values()))
    layers_list.insert(index, layer)
    return list2dictionary(layers_list, True, type)


def remove_layer(layers: dict, key: str) -> dict:
    """
    Removes a layer from a dictionary of layers, specified by its key.

    Parameters
    ----------
    layers : dict
        The dictionary of layers to remove the layer from.
    key : str
        The key of the layer to be removed.

    Returns
    -------
    layers : dict
        The updated dictionary of layers.
    """
    layers.pop(key)
    return layers
